"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import copy
import os
import sys
from array import array
from collections import namedtuple
from pathlib import Path
from typing import Dict, List, Tuple, Union

import imgui
import moderngl_window
import numpy as np
from moderngl_window import activate_context, get_local_window_cls
from PIL import Image
from tqdm import tqdm

from aitviewer.configuration import CONFIG as C
from aitviewer.remote.message import Message
from aitviewer.renderables.billboard import Billboard
from aitviewer.renderables.meshes import Meshes, VariableTopologyMeshes
from aitviewer.renderables.point_clouds import PointClouds
from aitviewer.renderer import Renderer, Viewport
from aitviewer.scene.camera import PinholeCamera, ViewerCamera
from aitviewer.scene.node import Node
from aitviewer.scene.scene import Scene
from aitviewer.server import ViewerServer
from aitviewer.shaders import clear_shader_cache
from aitviewer.streamables.streamable import Streamable
from aitviewer.utils import path
from aitviewer.utils.imgui_integration import ImGuiRenderer
from aitviewer.utils.perf_timer import PerfTimer
from aitviewer.utils.utils import get_video_paths, video_to_gif

MeshMouseIntersection = namedtuple(
    "MeshMouseIntersection",
    "node instance_id tri_id vert_id point_world point_local bc_coords",
)

if os.name == "nt":
    import ctypes

    try:
        # On windows we need to modify the current App User Model Id to make the taskbar icon of the viewer
        # the one that we set for the window and not the default python icon.
        #
        # For more details see:
        # https://stackoverflow.com/questions/1551605/how-to-set-applications-taskbar-icon-in-windows-7/1552105#1552105
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID("ait.viewer.window.1")
    except:
        pass

SHORTCUTS = {
    "SPACE": "Start/stop playing animation.",
    ".": "Go to next frame.",
    ",": "Go to previous frame.",
    "G": "Open a window to change frame by typing the frame number.",
    "X": "Center view on the selected object.",
    "O": "Enable/disable orthographic camera.",
    "T": "Show the camera target in the scene.",
    "B": "Show the camera trackball in the scene.",
    "C": "Save the camera position and orientation to disk.",
    "L": "Load the camera position and orientation from disk.",
    "K": "Lock the selection to the currently selected object.",
    "S": "Show/hide shadows.",
    "D": "Enabled/disable dark mode.",
    "P": "Save a screenshot to the the 'export/screenshots' directory.",
    "I": "Change the viewer mode to 'inspect'",
    "V": "Change the viewer mode to 'view'",
    "E": "If a mesh is selected, show the edges of the mesh.",
    "F": "If a mesh is selected, switch between flat and smooth shading.",
    "Z": "Show a debug visualization of the object IDs.",
    "ESC": "Exit the viewer.",
}


class Viewer(moderngl_window.WindowConfig):
    resource_dir = Path(__file__).parent / "shaders"
    size_mult = 1.0
    samples = 4
    if sys.platform == "darwin":
        gl_version = (4, 0)
    else:
        gl_version = (4, 5)
    window_type = None

    def __init__(
        self,
        title="aitviewer",
        size: Tuple[int, int] = None,
        samples: int = None,
        **kwargs,
    ):
        """
        Initializer.
        :param title: Window title
        :param size: Window size as (width, height) tuple, if None uses the size from the configuration file
        :param kwargs: kwargs.
        """
        # Window Setup (Following `moderngl_window.run_window_config`).
        if self.window_type is None:
            self.window_type = C.window_type

        # HACK: We use our own version of the PyQt5 windows to override
        # part of the initialization that crashes on Python >= 3.10.
        if self.window_type == "pyqt5":
            from aitviewer.utils.pyqt5_window import PyQt5Window

            base_window_cls = PyQt5Window
        else:
            base_window_cls = get_local_window_cls(self.window_type)

        # If no size is provided use the size from the configuration file.
        if size is None:
            size = C.window_width, C.window_height

        # Update nubmer of samples to use if specified as a parameter.
        if samples is not None:
            self.samples = samples

        # Calculate window size
        size = int(size[0] * self.size_mult), int(size[1] * self.size_mult)

        self.window = base_window_cls(
            title=title,
            size=size,
            fullscreen=C.fullscreen,
            resizable=C.resizable,
            gl_version=self.gl_version,
            aspect_ratio=None,  # Have to set this to None otherwise the window will enforce this aspect ratio.
            vsync=C.vsync,  # Set to False for some performance gains.
            samples=self.samples,
            cursor=True,
        )

        self.window_size = size
        self.window.print_context_info()
        activate_context(window=self.window)

        # Try to set the window icon.
        try:
            icon_path = os.path.join(os.path.dirname(__file__), "..", "assets", "aitviewer_icon.png")
            self.window.set_icon(icon_path=icon_path)
        except:
            pass

        self.timer = PerfTimer()
        self.ctx = self.window.ctx
        super().__init__(self.ctx, self.window, self.timer)

        # Create renderer
        self.renderer = Renderer(self.ctx, self.wnd, self.window_type)

        # Create GUI context
        self.imgui_ctx = imgui.create_context()
        self.imgui = ImGuiRenderer(self.wnd, self.window_type)
        self.imgui_user_interacting = False

        # Custom UI Font
        self.font_dir = Path(__file__).parent / "resources" / "fonts"
        self.fonts = imgui.get_io().fonts
        self.custom_font = self.fonts.add_font_from_file_ttf(os.path.join(self.font_dir, "Custom.ttf"), 15)
        self.imgui.refresh_font_texture()

        self.modes = {
            "view": {"title": " View", "shortcut": "V"},
            "inspect": {"title": " Inspect", "shortcut": "I"},
        }

        self.gui_controls = {
            "menu": self.gui_menu,
            "scene": self.gui_scene,
            "playback": self.gui_playback,
            "inspect": self.gui_inspect,
            "shortcuts": self.gui_shortcuts,
            "exit": self.gui_exit,
            "go_to_frame": self.gui_go_to_frame,
        }

        # Initialize viewer
        self.scene = None
        self.reset()

        # Key Shortcuts
        self._exit_key = self.wnd.keys.ESCAPE
        self._pause_key = self.wnd.keys.SPACE
        self._next_frame_key = self.wnd.keys.PERIOD
        self._previous_frame_key = self.wnd.keys.COMMA
        self._shadow_key = self.wnd.keys.S
        self._orthographic_camera_key = self.wnd.keys.O
        self._center_view_on_selection_key = self.wnd.keys.X
        self._dark_mode_key = self.wnd.keys.D
        self._screenshot_key = self.wnd.keys.P
        self._middle_mouse_button = 3  # middle
        self._right_mouse_button = 2  # right
        self._left_mouse_button = 1  # left
        self._save_cam_key = self.wnd.keys.C
        self._load_cam_key = self.wnd.keys.L
        self._show_camera_target_key = self.wnd.keys.T
        self._show_camera_trackball_key = self.wnd.keys.B
        self._visualize_key = self.wnd.keys.Z
        self._lock_selection_key = self.wnd.keys.K
        self._mode_inspect_key = self.wnd.keys.I
        self._mode_view_key = self.wnd.keys.V
        self._go_to_frame_key = self.wnd.keys.G
        self._shortcut_names = {
            self.wnd.keys.SPACE: "Space",
            self.wnd.keys.B: "B",
            self.wnd.keys.C: "C",
            self.wnd.keys.D: "D",
            self.wnd.keys.I: "I",
            self.wnd.keys.L: "L",
            self.wnd.keys.K: "K",
            self.wnd.keys.O: "O",
            self.wnd.keys.P: "P",
            self.wnd.keys.S: "S",
            self.wnd.keys.T: "T",
            self.wnd.keys.X: "X",
            self.wnd.keys.G: "G",
            self.wnd.keys.Z: "Z",
        }

        # Disable exit on escape key
        self.window.exit_key = None

        # GUI
        self._render_gui = True
        self._exit_popup_open = False
        self._screenshot_popup_open = False
        self._screenshot_popup_just_opened = False
        self._export_usd_popup_open = False
        self._export_usd_popup_just_opened = False
        self._go_to_frame_popup_open = False
        self._go_to_frame_string = ""
        self._show_shortcuts_window = False
        self._mouse_position = (0, 0)
        self._modal_focus_count = 0

        self.server = None
        if C.server_enabled:
            self.server = ViewerServer(self, C.server_port)

    # noinspection PyAttributeOutsideInit
    def reset(self):
        if self.scene is not None:
            self.scene.release()

        # Setup scene
        self.scene = Scene()
        self.scene.camera = ViewerCamera(45.0)
        self.scene.custom_font = self.custom_font

        # Settings
        self.run_animations = C.run_animations
        self.playback_fps = C.playback_fps
        self.playback_without_skipping = False
        self.shadows_enabled = C.shadows_enabled
        self.auto_set_floor = C.auto_set_floor
        self.auto_set_camera_target = C.auto_set_camera_target
        self.backface_culling = C.backface_culling
        self.scene.light_mode = "dark" if C.dark_mode else "default"
        self.lock_selection = False
        self.visualize = False

        self._pan_camera = False
        self._rotate_camera = False
        self._past_frametimes = np.zeros([60]) - 1.0
        self._last_frame_rendered_at = 0

        # Outline colors
        self.outline_color = (0.3, 0.7, 1.0, 1)
        self.light_outline_color = (0.4, 0.4, 0.4, 1)
        self.selected_outline_color = (1.0, 0.86, 0.35, 1.0)

        # Mouse mesh intersection in inspect mode.
        self.mmi = None

        # Mouse selection
        self._move_threshold = 3
        self._mouse_moved = False
        self._mouse_down_position = np.array([0, 0])

        # Export settings
        self.export_animation = True
        self.export_animation_range = [0, -1]
        self.export_duration = 10
        self.export_format = "mp4"
        self.export_rotate_camera = False
        self.export_rotation_degrees = 360
        self.export_fps = self.playback_fps
        self.export_scale_factor = 1.0
        self.export_transparent = False
        self.export_quality = "medium"

        # Screenshot settings
        self.screenshot_transparent = False
        self.screenshot_name = None

        # Export usd settings
        self.export_usd_name = None
        self.export_usd_directory = False

        # Setup viewports
        self.viewports: List[Viewport] = []
        self._viewport_mode = None
        self.viewport_mode = "single"

        # Set the mode once the viewer has been completely initialized
        self.selected_mode = "view"

    @property
    def render_gui(self):
        return self._render_gui

    @render_gui.setter
    def render_gui(self, value):
        self._render_gui = value

    def _resize_viewports(self):
        """
        Computes the size of the viewports depending on the current mode.
        Uses the window buffer size to update the extents of all existing viewports.
        """
        mode = self.viewport_mode

        w, h = self.wnd.buffer_size
        if mode == "single":
            self.viewports[0].extents = [0, 0, w, h]
        elif mode == "split_v":
            self.viewports[0].extents = [0, 0, w // 2 - 1, h]
            self.viewports[1].extents = [w // 2 + 1, 0, w - w // 2 - 1, h]
        elif mode == "split_h":
            self.viewports[0].extents = [0, h // 2 + 1, w, h - h // 2 - 1]
            self.viewports[1].extents = [0, 0, w, h // 2 - 1]
        elif mode == "split_vh":
            self.viewports[0].extents = [0, h // 2 + 1, w // 2 - 1, h - h // 2 - 1]
            self.viewports[1].extents = [w // 2 + 1, h // 2 + 1, w - w // 2 - 1, h - h // 2 - 1]
            self.viewports[2].extents = [0, 0, w // 2 - 1, h // 2 - 1]
            self.viewports[3].extents = [w // 2 + 1, 0, w - w // 2 - 1, h // 2 - 1]
        else:
            raise ValueError(f"Invalid viewport mode: {mode}")

    def set_ortho_grid_viewports(self, invert=(False, False, False)):
        """
        Sets the viewports in a 2x2 grid with a viewport looking at the scene with an
        orthographic projection from each of the main axes.

        :param invert: a tuple of 3 booleans, for the x, y and z axis respectively.
        If False the camera for the respective axis is placed at the positive side looking
        towards the negative direction, and opposite otherwise.
        """
        self.viewport_mode = "split_vh"

        bounds = self.scene.bounds_without_floor
        center = bounds.mean(-1)
        size = bounds[:, 1] - bounds[:, 0]

        axis = [
            np.array([1, 0, 0]),
            np.array([0, 0, 1]),
            np.array([0, 1, 0]),
        ]
        for i in range(3):
            if invert[i]:
                axis[i] = -axis[i]

        up = [
            np.array([0, 1, 0]),
            np.array([0, 1, 0]),
            np.array([0, 0, 1]),
        ]

        ar = self.window.width / self.window.height

        def compute_ortho_size(w, h):
            if w / h > ar:
                # Limiting size is width, convert to limit on height.
                size = w / ar
            else:
                # Limiting size is height, return directly
                size = h
            return size * 0.55

        ortho_size = [
            compute_ortho_size(size[2], size[1]),
            compute_ortho_size(size[0], size[1]),
            compute_ortho_size(size[0], size[2]),
        ]

        for i, v in enumerate(self.viewports[1:]):
            v.camera = ViewerCamera(orthographic=ortho_size[i])
            v.camera.target = center
            v.camera.position = center + size[i] * axis[i]
            v.camera.up = up[i]

    @property
    def viewport_mode(self):
        """Getter for the current viewport mode"""
        return self._viewport_mode

    @viewport_mode.setter
    def viewport_mode(self, mode):
        """
        Setter for the current viewport mode, valid values are: "single", "split_v",  "split_h" and  "split_hv"
        which split the window in a 1x1, 1x2, 2x1 and 2x2 grid respectively.
        """
        if mode == self._viewport_mode:
            return

        self._viewport_mode = mode
        num_viewports = len(self.viewports)
        if mode == "single":
            new_num_viewports = 1
        elif mode == "split_v":
            new_num_viewports = 2
        elif mode == "split_h":
            new_num_viewports = 2
        elif mode == "split_vh":
            new_num_viewports = 4
        else:
            raise ValueError(f"Invalid viewport mode: {mode}")

        if num_viewports > new_num_viewports:
            # If the camera of one of the viewports that we are removing is selected
            # we just deselct it and remove it from the GUI, since it won't exist anymore.
            for v in self.viewports[new_num_viewports:]:
                if self.scene.selected_object == v.camera and isinstance(v.camera, ViewerCamera):
                    self.scene.select(None)
                    self.scene.gui_selected_object = None
            # Remove extra viewports.
            self.viewports = self.viewports[:new_num_viewports]
        elif num_viewports < new_num_viewports:
            for _ in range(num_viewports, new_num_viewports):
                camera = self.scene.camera
                if isinstance(self.scene.camera, ViewerCamera):
                    camera = camera.copy()
                self.viewports.append(Viewport([], camera))
        self._resize_viewports()

    def _init_scene(self):
        self.scene.make_renderable(self.ctx)
        if self.auto_set_floor:
            self.scene.auto_set_floor()
            self.scene.backface_culling = self.backface_culling
        if self.auto_set_camera_target:
            self.scene.auto_set_camera_target()

    def get_node_by_remote_uid(self, remote_uid: int, client: Tuple[str, str]):
        """
        Returns the Node corresponding to the remote uid and client passed in.

        :param remote_uid: the remote uid to look up.
        :param client: the client that created the node, this is the value of the 'client'
            parameter that was passed to process_message() when the message was received.
        :return: Node corresponding to the remote uid.
        """
        return self.server.get_node_by_remote_uid(remote_uid, client)

    def process_message(self, type: Message, remote_uid: int, args: list, kwargs: dict, client: Tuple[str, str]):
        """
        Default processing of messages received by the viewer.

        This method is called every time a new message has to be processed.
        It can be overriden to intercept messages before they are sent to
        the viewer and to add custom functionality.

        :param type: an integer id that represents the type of the message.
        :param remote_uid: the remote id of the node that this message refers to.
            This is an id generated by the client to reference nodes that
            it created. The viewer keeps track of the mapping from remote_uid
            to the local uid of the node.
            Use 'viewer.get_node_by_remote_uid(remote_uid)' to get the
            Node corresponding to this id.
        :param args: positional arguments received with the message.
        :param kwargs: keyword arguments received with the message.
        :param client: a tuple (ip, port) describing the address of the client
            that sent this message.
        """
        try:
            self.server.process_message(type, remote_uid, args, kwargs, client)
        except Exception as e:
            print(f"Exception while processing mesage: type = {type}, remote_uid = {remote_uid}:\n{e}")

    def send_message(self, msg, client: Tuple[str, str] = None):
        """
        Send a message to a single client or to all connected clients.

        :param msg: a python object that is serialized with pickle and sent to the client.
        :param client: a tuple (host, port) representing the client to which to send the message,
            if None the message is sent to all connected clients.
        """
        if self.server is not None:
            self.server.send_message(msg, client)

    def run(self, *args, log=True):
        """
        Enter a blocking visualization loop. This is built following `moderngl_window.run_window_config`.
        :param args: The arguments passed to `config_cls` constructor.
        :param log: Whether to log to the console.
        """
        self._init_scene()

        self.export_animation_range[-1] = self.scene.n_frames - 1

        self.timer.start()
        self._last_frame_rendered_at = self.timer.time

        while not self.window.is_closing:
            if self.server is not None:
                self.server.process_messages()

            current_time, delta = self.timer.next_frame()

            self.window.clear()
            self.window.render(current_time, delta)
            self.window.swap_buffers()
        _, duration = self.timer.stop()
        self.on_close()
        imgui.destroy_context(self.imgui_ctx)
        # Necessary for pyglet window, otherwise the window is not closed.
        self.window.close()
        self.window.destroy()
        if duration > 0 and log:
            print("Duration: {0:.2f}s @ {1:.2f} FPS".format(duration, self.window.frames / duration))

    def render(self, time, frame_time, export=False, transparent_background=False):
        """The main drawing function."""

        if self.run_animations:
            # Compute number of frames to advance by.
            frames = (int)((time - self._last_frame_rendered_at) * self.playback_fps)
            if frames > 0:
                if self.playback_without_skipping:
                    frames = 1
                self.scene.current_frame_id = (self.scene.current_frame_id + frames) % self.scene.n_frames
                self._last_frame_rendered_at += frames * (1.0 / self.playback_fps)

        # Update camera matrices that will be used for rendering
        if isinstance(self.scene.camera, ViewerCamera):
            self.scene.camera.update_animation(frame_time)

        # Update streamable captures.
        if not export:
            self.streamable_capture()

        # Render.
        self.renderer.render(
            self.scene,
            viewports=self.viewports,
            outline_color=self.outline_color,
            light_outline_color=self.light_outline_color,
            selected_outline_color=self.selected_outline_color,
            export=export,
            transparent_background=transparent_background,
            window_size=self.window_size,
            shadows_enabled=self.shadows_enabled,
            visualize=self.visualize,
        )

        if not export:
            # FPS accounting.
            self._past_frametimes[:-1] = self._past_frametimes[1:]
            self._past_frametimes[-1] = frame_time

            # Render the UI components.
            self.gui()

    def streamable_capture(self):
        # Collect all streamable nodes
        rs = self.scene.collect_nodes(obj_type=Streamable)
        for r in rs:
            r.capture()

    def prevent_background_interactions(self):
        """Prevent background interactions when hovering over any imgui window."""
        self.imgui_user_interacting = self.imgui.io.want_capture_mouse

    def toggle_animation(self, run: bool):
        self.run_animations = run
        if self.run_animations:
            self._last_frame_rendered_at = self.timer.time

    def reset_camera(self, viewport=None):
        if viewport is None:
            viewport = self.viewports[0]
        viewport.reset_camera()
        if viewport == self.viewports[0]:
            self.scene.camera = self.viewports[0].camera

    def set_temp_camera(self, camera, viewport=None):
        if viewport is None:
            viewport = self.viewports[0]
        viewport.set_temp_camera(camera)
        if viewport == self.viewports[0]:
            self.scene.camera = self.viewports[0].camera

    def lock_to_node(self, node: Node, relative_position, smooth_sigma=None, viewport=None):
        """
        Create and return a PinholeCamera that follows a node, the target of the camera is the center
        of the node at each frame and the camera is positioned with a constant offset (relative_position)
        from its target. See aitviewer.utils.path.lock_to_node for more details about parameters.
        The camera is set as the current viewer camera.
        """
        pos, tar = path.lock_to_node(node, relative_position, smooth_sigma=smooth_sigma)
        cam = PinholeCamera(pos, tar, self.window_size[0], self.window_size[1], viewer=self)
        self.scene.add(cam)
        self.set_temp_camera(cam, viewport)
        return cam

    def get_viewport_at_position(self, x: int, y: int) -> Union[Viewport, None]:
        x, y = self._mouse_to_buffer(x, y)
        for v in self.viewports:
            if v.contains(x, y):
                return v
        return None

    def gui(self):
        imgui.new_frame()

        # Add viewport separators to draw list.
        w, h = self.wnd.buffer_size
        draw = imgui.get_background_draw_list()
        c = 36 / 255
        color = imgui.get_color_u32_rgba(c, c, c, 1.0)
        if self.viewport_mode == "split_v" or self.viewport_mode == "split_vh":
            draw.add_rect_filled(w // 2 - 1, 0, w // 2 + 1, h, color)
        if self.viewport_mode == "split_h" or self.viewport_mode == "split_vh":
            draw.add_rect_filled(0, h // 2 - 1, w, h // 2 + 1, color)

        # Create a context menu when right clicking on the background.
        if (
            not any([imgui.is_window_hovered(), imgui.is_any_item_hovered()])
            and imgui.is_mouse_released(button=1)
            and not self._mouse_moved
        ):
            x, y = imgui.get_io().mouse_pos
            # Select the object under the cursor.
            if self.select_object(x, y) or not isinstance(self.scene.selected_object, Node):
                self._context_menu_position = (x, y)
                imgui.open_popup("Context Menu")

        # Draw the context menu for the selected object.
        if imgui.begin_popup("Context Menu"):
            if self.scene.selected_object is None or not isinstance(self.scene.selected_object, Node):
                imgui.close_current_popup()
            else:
                self.scene.selected_object.gui_context_menu(imgui, *self._context_menu_position)
            imgui.end_popup()

        # Reset user interacting state.
        self.imgui_user_interacting = False

        if self.render_gui:
            # Render user controls.
            for gc in self.gui_controls.values():
                gc()
        else:
            # If gui is disabled only render the go to frame window.
            self.gui_go_to_frame()

        # Contains live examples of all possible displays/controls - useful for browsing for new components
        # imgui.show_test_window()

        imgui.render()
        self.imgui.render(imgui.get_draw_data())

        self.prevent_background_interactions()

    def gui_scene(self):
        # Render scene GUI
        imgui.set_next_window_position(50, 50, imgui.FIRST_USE_EVER)
        imgui.set_next_window_size(self.window_size[0] * 0.25, self.window_size[1] * 0.7, imgui.FIRST_USE_EVER)
        expanded, _ = imgui.begin("Editor", None)
        if expanded:
            self.scene.gui_editor(imgui, self.viewports, self.viewport_mode)
        imgui.end()

    def gui_menu(self):
        clicked_export = False
        clicked_screenshot = False
        clicked_export_usd = False

        if imgui.begin_main_menu_bar():
            if imgui.begin_menu("File", True):
                clicked_quit, selected_quit = imgui.menu_item("Quit", "Cmd+Q", False, True)
                if clicked_quit:
                    exit(1)

                clicked_export, selected_export = imgui.menu_item("Save as video..", None, False, True)

                clicked_screenshot, selected_screenshot = imgui.menu_item(
                    "Screenshot",
                    self._shortcut_names[self._screenshot_key],
                    False,
                    True,
                )

                clicked_export_usd, _ = imgui.menu_item("Save as USD", None, False, True)
                imgui.end_menu()

            if imgui.begin_menu("View", True):
                if imgui.begin_menu("Light modes"):
                    _, default = imgui.menu_item("Default", None, self.scene.light_mode == "default")
                    if default:
                        self.scene.light_mode = "default"

                    _, dark = imgui.menu_item(
                        "Dark",
                        self._shortcut_names[self._dark_mode_key],
                        self.scene.light_mode == "dark",
                    )
                    if dark:
                        self.scene.light_mode = "dark"

                    _, diffuse = imgui.menu_item("Diffuse", None, self.scene.light_mode == "diffuse")
                    if diffuse:
                        self.scene.light_mode = "diffuse"
                    imgui.end_menu()

                _, self.shadows_enabled = imgui.menu_item(
                    "Render Shadows",
                    self._shortcut_names[self._shadow_key],
                    self.shadows_enabled,
                    True,
                )

                _, self.lock_selection = imgui.menu_item(
                    "Lock selection",
                    self._shortcut_names[self._lock_selection_key],
                    self.lock_selection,
                    True,
                )

                if imgui.begin_menu("Viewports"):

                    def menu_entry(text, name):
                        if imgui.menu_item(text, None, self.viewport_mode == name)[1]:
                            self.viewport_mode = name

                    menu_entry("Single", "single")
                    menu_entry("Split vertical", "split_v")
                    menu_entry("Split horizontal", "split_h")
                    menu_entry("Split both", "split_vh")
                    imgui.separator()
                    if imgui.menu_item("Ortho grid +", None)[1]:
                        self.set_ortho_grid_viewports()
                    if imgui.menu_item("Ortho grid -", None)[1]:
                        self.set_ortho_grid_viewports((True, True, True))

                    imgui.end_menu()

                _, self.render_gui = imgui.menu_item("Render GUI", None, self.render_gui, True)

                imgui.end_menu()

            if imgui.begin_menu("Camera", True):
                _, self.scene.camera_target.enabled = imgui.menu_item(
                    "Show camera target",
                    self._shortcut_names[self._show_camera_target_key],
                    self.scene.camera_target.enabled,
                    True,
                )

                _, self.scene.trackball.enabled = imgui.menu_item(
                    "Show camera trackball",
                    self._shortcut_names[self._show_camera_trackball_key],
                    self.scene.trackball.enabled,
                    True,
                )

                clicked, _ = imgui.menu_item(
                    "Center view on selection",
                    self._shortcut_names[self._center_view_on_selection_key],
                    False,
                    isinstance(self.scene.selected_object, Node),
                )
                if clicked:
                    self.center_view_on_selection()

                # TODO (@Dario): correct to only use self.viewports[0] here?
                is_ortho = False if self.viewports[0].using_temp_camera else self.scene.camera.is_ortho
                _, is_ortho = imgui.menu_item(
                    "Orthographic Camera",
                    self._shortcut_names[self._orthographic_camera_key],
                    is_ortho,
                    True,
                )
                # TODO (@Dario): correct to only use self.viewports[0] here?
                if is_ortho and self.viewports[0].using_temp_camera:
                    self.reset_camera()
                self.scene.camera.is_ortho = is_ortho

                # TODO (@Dario)
                if not self.viewports[0].using_temp_camera and imgui.begin_menu("Control modes"):

                    def mode(name, mode):
                        selected = imgui.menu_item(name, None, self.scene.camera.control_mode == mode)[1]
                        if selected:
                            self.reset_camera()
                            self.scene.camera.control_mode = mode

                    mode("Turntable", "turntable")
                    mode("Trackball", "trackball")
                    mode("First Person", "first_person")
                    imgui.end_menu()

                clicked_save_cam, selected_save_cam = imgui.menu_item(
                    "Save Camera", self._shortcut_names[self._save_cam_key], False, True
                )
                if clicked_save_cam:
                    self.reset_camera()
                    self.scene.camera.save_cam()

                clicked_load_cam, selected_load_cam = imgui.menu_item(
                    "Load Camera", self._shortcut_names[self._load_cam_key], False, True
                )
                if clicked_load_cam:
                    self.reset_camera()
                    self.scene.camera.load_cam()

                imgui.end_menu()

            if imgui.begin_menu("Mode", True):
                for id, mode in self.modes.items():
                    mode_clicked, _ = imgui.menu_item(mode["title"], mode["shortcut"], id == self.selected_mode, True)
                    if mode_clicked:
                        self.selected_mode = id

                imgui.end_menu()

            if imgui.begin_menu("Help", True):
                clicked, self._show_shortcuts_window = imgui.menu_item(
                    "Keyboard shortcuts", None, self._show_shortcuts_window
                )
                imgui.end_menu()

            if imgui.begin_menu("Debug", True):
                _, self.visualize = imgui.menu_item(
                    "Visualize debug texture",
                    self._shortcut_names[self._visualize_key],
                    self.visualize,
                    True,
                )

                imgui.end_menu()

            if self.server is not None:
                if imgui.begin_menu("Server", True):
                    imgui.text("Connected clients:")
                    imgui.separator()
                    for c in self.server.connections.keys():
                        imgui.text(f"{c[0]}:{c[1]}")
                    imgui.end_menu()

            imgui.end_main_menu_bar()

        if clicked_export:
            imgui.open_popup("Export Video")
            self.export_fps = self.playback_fps
            self.toggle_animation(False)

        if clicked_screenshot:
            self._screenshot_popup_just_opened = True

        if clicked_export_usd:
            self._export_usd_popup_just_opened = True

        self.gui_export_video()
        self.gui_screenshot()
        self.gui_export_usd()

    def gui_export_video(self):
        imgui.set_next_window_size(570, 0)
        if imgui.begin_popup_modal("Export Video", flags=imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_MOVE)[0]:
            if self.scene.n_frames == 1:
                imgui.push_style_var(imgui.STYLE_ALPHA, 0.2)
                self.export_animation = False

            if imgui.radio_button("Animation", self.export_animation):
                self.export_animation = True

            if self.scene.n_frames == 1:
                imgui.pop_style_var()
                self.export_animation = False

            imgui.same_line(spacing=15)
            if imgui.radio_button("360 shot", not self.export_animation):
                self.export_animation = False

            if self.export_animation:
                _, animation_range = imgui.drag_int2(
                    "Animation range",
                    *self.export_animation_range,
                    min_value=0,
                    max_value=self.scene.n_frames - 1,
                )
                if animation_range[0] != self.export_animation_range[0]:
                    self.export_animation_range[0] = animation_range[0]
                elif animation_range[-1] != self.export_animation_range[1]:
                    self.export_animation_range[1] = animation_range[1]

                _, self.playback_fps = imgui.drag_float(
                    "Playback fps",
                    self.playback_fps,
                    0.1,
                    min_value=1.0,
                    max_value=120.0,
                    format="%.1f",
                )
                imgui.same_line(spacing=10)
                speedup = self.playback_fps / self.scene.fps
                imgui.text(f"({speedup:.2f}x speed)")
            else:
                _, self.scene.current_frame_id = imgui.slider_int(
                    "Frame",
                    self.scene.current_frame_id,
                    min_value=0,
                    max_value=self.scene.n_frames - 1,
                )
                _, self.export_duration = imgui.drag_float(
                    "Duration (s)",
                    self.export_duration,
                    min_value=0.1,
                    max_value=10000.0,
                    change_speed=0.05,
                    format="%.1f",
                )
                duration = self.export_duration

            if self.export_animation:
                if isinstance(self.scene.camera, ViewerCamera):
                    _, self.export_rotate_camera = imgui.checkbox("Rotate camera", self.export_rotate_camera)
                else:
                    imgui.push_style_var(imgui.STYLE_ALPHA, 0.2)
                    imgui.checkbox("Rotate camera (only available for ViewerCamera)", False)
                    imgui.pop_style_var(1)

            if not self.export_animation or self.export_rotate_camera:
                _, self.export_rotation_degrees = imgui.drag_int(
                    "Rotation angle (degrees)", self.export_rotation_degrees
                )

            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            # Output settings.
            imgui.text("Output")

            imgui.spacing()
            imgui.text("Format:")
            imgui.same_line()
            if imgui.radio_button("MP4", self.export_format == "mp4"):
                self.export_format = "mp4"
            imgui.same_line(spacing=15)
            if imgui.radio_button("WEBM", self.export_format == "webm"):
                self.export_format = "webm"
            imgui.same_line(spacing=15)
            if imgui.radio_button("GIF", self.export_format == "gif"):
                self.export_format = "gif"

            imgui.spacing()
            if self.export_format != "webm":
                imgui.push_style_var(imgui.STYLE_ALPHA, 0.2)
                self.export_transparent = False

            _, self.export_transparent = imgui.checkbox("Transparent background", self.export_transparent)
            if self.export_format != "webm":
                imgui.pop_style_var()
                self.export_transparent = False

            if imgui.is_item_hovered():
                imgui.begin_tooltip()
                imgui.text("Available only for WEBM format")
                imgui.end_tooltip()

            if self.export_format == "gif":
                max_output_fps = 30.0
            else:
                max_output_fps = 120.0
            self.export_fps = min(self.export_fps, max_output_fps)
            _, self.export_fps = imgui.drag_float(
                "fps",
                self.export_fps,
                0.1,
                min_value=1.0,
                max_value=max_output_fps,
                format="%.1f",
            )
            imgui.same_line(position=440)
            if imgui.button("1x##fps", width=35):
                self.export_fps = self.playback_fps
            imgui.same_line()
            if imgui.button("1/2x##fps", width=35):
                self.export_fps = self.playback_fps / 2
            imgui.same_line()
            if imgui.button("1/4x##fps", width=35):
                self.export_fps = self.playback_fps / 4

            imgui.spacing()
            imgui.spacing()
            imgui.text(
                f"Resolution: [{int(self.window_size[0] * self.export_scale_factor)}x{int(self.window_size[1] * self.export_scale_factor)}]"
            )
            _, self.export_scale_factor = imgui.drag_float(
                "Scale",
                self.export_scale_factor,
                min_value=0.01,
                max_value=1.0,
                change_speed=0.005,
                format="%.2f",
            )

            imgui.same_line(position=440)
            if imgui.button("1x##scale", width=35):
                self.export_scale_factor = 1.0
            imgui.same_line()
            if imgui.button("1/2x##scale", width=35):
                self.export_scale_factor = 0.5
            imgui.same_line()
            if imgui.button("1/4x##scale", width=35):
                self.export_scale_factor = 0.25

            imgui.spacing()
            if self.export_format == "mp4":
                imgui.text("Quality: ")
                imgui.same_line()
                if imgui.radio_button("high", self.export_quality == "high"):
                    self.export_quality = "high"
                imgui.same_line()
                if imgui.radio_button("medium", self.export_quality == "medium"):
                    self.export_quality = "medium"
                imgui.same_line()
                if imgui.radio_button("low", self.export_quality == "low"):
                    self.export_quality = "low"

            if self.export_animation:
                duration = (animation_range[1] - animation_range[0] + 1) / self.playback_fps
                # Compute exact number of frames if playback fps is an exact multiple of export fps.
                if np.fmod(self.playback_fps, self.export_fps) < 0.1:
                    playback_count = int(np.round(self.playback_fps / self.export_fps))
                    frames = (animation_range[1] - animation_range[0] + 1) // playback_count
                else:
                    frames = int(np.ceil(duration * self.export_fps))
            else:
                frames = int(np.ceil(duration * self.export_fps))

            imgui.spacing()
            if frames > 0:
                imgui.text(f"Duration: {duration:.2f}s ({frames} frames @ {self.export_fps:.2f}fps)")
            else:
                if self.export_animation:
                    imgui.text(f"Error: Animation range is empty")
                else:
                    imgui.text(f"Error: Duration is 0 seconds")
            imgui.spacing()

            # Draw a cancel and exit button on the same line using the available space
            button_width = (imgui.get_content_region_available()[0] - imgui.get_style().item_spacing[0]) * 0.5

            # Style the cancel with a grey color
            imgui.push_style_color(imgui.COLOR_BUTTON, 0.5, 0.5, 0.5, 1.0)
            imgui.push_style_color(imgui.COLOR_BUTTON_ACTIVE, 0.6, 0.6, 0.6, 1.0)
            imgui.push_style_color(imgui.COLOR_BUTTON_HOVERED, 0.7, 0.7, 0.7, 1.0)

            if imgui.button("Cancel", width=button_width):
                imgui.close_current_popup()

            imgui.pop_style_color()
            imgui.pop_style_color()
            imgui.pop_style_color()

            imgui.same_line()

            if frames <= 0:
                imgui.push_style_var(imgui.STYLE_ALPHA, 0.2)

            if imgui.button("Export", button_width) and frames >= 0:
                imgui.close_current_popup()
                self.export_video(
                    os.path.join(
                        C.export_dir,
                        "videos",
                        f"{self.window.title}.{self.export_format}",
                    ),
                    animation=self.export_animation,
                    animation_range=self.export_animation_range,
                    duration=self.export_duration,
                    frame=self.scene.current_frame_id,
                    output_fps=self.export_fps,
                    rotate_camera=not self.export_animation or self.export_rotate_camera,
                    rotation_degrees=self.export_rotation_degrees,
                    scale_factor=self.export_scale_factor,
                    transparent=self.export_transparent,
                    quality=self.export_quality,
                )

            if frames <= 0:
                imgui.pop_style_var()

            imgui.end_popup()

    def gui_playback(self):
        """GUI to control playback settings."""
        imgui.set_next_window_position(50, 100 + self.window_size[1] * 0.7, imgui.FIRST_USE_EVER)
        imgui.set_next_window_size(self.window_size[0] * 0.4, self.window_size[1] * 0.15, imgui.FIRST_USE_EVER)
        expanded, _ = imgui.begin("Playback", None)
        if expanded:
            u, run_animations = imgui.checkbox(
                "Run animations [{}]".format(self._shortcut_names[self._pause_key]),
                self.run_animations,
            )
            if u:
                self.toggle_animation(run_animations)

            # Plot FPS
            frametime_avg = np.mean(self._past_frametimes[self._past_frametimes > 0.0])
            fps_avg = 1 / frametime_avg
            ms_avg = frametime_avg * 1000.0
            ms_last = self._past_frametimes[-1] * 1000.0

            imgui.plot_lines(
                "Internal {:.1f} fps @ {:.2f} ms/frame [{:.2f}ms]".format(fps_avg, ms_avg, ms_last),
                array("f", (1.0 / self._past_frametimes).tolist()),
                scale_min=0,
                scale_max=100.0,
                graph_size=(100, 20),
            )

            _, self.playback_fps = imgui.drag_float(
                f"Playback fps",
                self.playback_fps,
                0.1,
                min_value=1.0,
                max_value=120.0,
                format="%.1f",
            )
            imgui.same_line(spacing=10)
            speedup = self.playback_fps / self.scene.fps
            imgui.text(f"({speedup:.2f}x speed)")

            # Sequence Control
            # For simplicity, we allow the global sequence slider to only go as far as the shortest known sequence.
            n_frames = self.scene.n_frames

            _, self.scene.current_frame_id = imgui.slider_int(
                "Frame##r_global_seq_control",
                self.scene.current_frame_id,
                min_value=0,
                max_value=n_frames - 1,
            )
            self.prevent_background_interactions()
        imgui.end()

    def gui_shortcuts(self):
        if self._show_shortcuts_window:
            imgui.set_next_window_position(self.window_size[0] * 0.6, 200, imgui.FIRST_USE_EVER)
            imgui.set_next_window_size(self.window_size[0] * 0.35, 350, imgui.FIRST_USE_EVER)
            expanded, self._show_shortcuts_window = imgui.begin("Keyboard shortcuts", self._show_shortcuts_window)
            if expanded:
                for k, v in SHORTCUTS.items():
                    imgui.bullet_text(f"{k:5} - {v}")
            imgui.end()

    def gui_inspect(self):
        """GUI to control playback settings."""
        if self.selected_mode == "inspect":
            imgui.set_next_window_position(self.window_size[0] * 0.6, 50, imgui.FIRST_USE_EVER)
            imgui.set_next_window_size(self.window_size[0] * 0.35, 140, imgui.FIRST_USE_EVER)
            expanded, _ = imgui.begin("Inspect", None)
            if expanded:
                if self.mmi is not None:
                    for k, v in zip(self.mmi._fields, self.mmi):
                        imgui.text("{}: {}".format(k, v))

                self.prevent_background_interactions()
            imgui.end()

    def gui_go_to_frame(self):
        if self._go_to_frame_popup_open:
            imgui.open_popup("Go to frame##go-to-frame-popup")
            self._go_to_frame_string = str(self.scene.current_frame_id)

        imgui.set_next_window_size(300, 0)
        if imgui.begin_popup_modal(
            "Go to frame##go-to-frame-popup",
            flags=imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_MOVE | imgui.WINDOW_NO_TITLE_BAR,
        )[0]:
            if self._go_to_frame_popup_open:
                imgui.set_keyboard_focus_here()
                u, self._go_to_frame_string = imgui.input_text(
                    "Go to frame",
                    self._go_to_frame_string,
                    64,
                    imgui.INPUT_TEXT_CHARS_DECIMAL
                    | imgui.INPUT_TEXT_ENTER_RETURNS_TRUE
                    | imgui.INPUT_TEXT_AUTO_SELECT_ALL,
                )
                if u:
                    try:
                        frame_id = int(self._go_to_frame_string)
                    except:
                        frame_id = -1
                        pass
                    if frame_id >= 0:
                        self.scene.current_frame_id = frame_id
                    self._go_to_frame_popup_open = False
                    imgui.close_current_popup()
            else:
                imgui.close_current_popup()
            imgui.end_popup()

    def gui_screenshot(self):
        if self._screenshot_popup_just_opened:
            self._screenshot_popup_just_opened = False
            self._screenshot_popup_open = True
            self.screenshot_name = None
            self._modal_focus_count = 2
            self.toggle_animation(False)
            imgui.open_popup("Screenshot##screenshot-popup")

        imgui.set_next_window_size(250, 0)
        if imgui.begin_popup_modal(
            "Screenshot##screenshot-popup",
            flags=imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_MOVE,
        )[0]:
            if self._screenshot_popup_open:
                _, self.screenshot_transparent = imgui.checkbox("Transparent background", self.screenshot_transparent)
                if self.screenshot_name is None:
                    self.screenshot_name = "frame_{:0>6}.png".format(self.scene.current_frame_id)

                # HACK: we need to set the focus twice when the modal is first opened for it to take effect
                if self._modal_focus_count > 0:
                    self._modal_focus_count -= 1
                    imgui.set_keyboard_focus_here()
                _, self.screenshot_name = imgui.input_text(
                    "File name",
                    self.screenshot_name,
                    64,
                    imgui.INPUT_TEXT_AUTO_SELECT_ALL,
                )
                imgui.spacing()

                button_width = (imgui.get_content_region_available()[0] - imgui.get_style().item_spacing[0]) * 0.5

                # Style the cancel with a grey color
                imgui.push_style_color(imgui.COLOR_BUTTON, 0.5, 0.5, 0.5, 1.0)
                imgui.push_style_color(imgui.COLOR_BUTTON_ACTIVE, 0.6, 0.6, 0.6, 1.0)
                imgui.push_style_color(imgui.COLOR_BUTTON_HOVERED, 0.7, 0.7, 0.7, 1.0)

                if imgui.button("cancel", width=button_width):
                    imgui.close_current_popup()
                    self._screenshot_popup_open = False

                imgui.pop_style_color()
                imgui.pop_style_color()
                imgui.pop_style_color()

                imgui.same_line()
                if imgui.button("save", button_width):
                    if self.screenshot_name:
                        self.take_screenshot(self.screenshot_name, self.screenshot_transparent)
                    imgui.close_current_popup()
                    self._screenshot_popup_open = False

            else:
                imgui.close_current_popup()
            imgui.end_popup()

    def gui_export_usd(self):
        if self._export_usd_popup_just_opened:
            self._export_usd_popup_just_opened = False
            self._export_usd_popup_open = True
            self.export_usd_name = None
            self._modal_focus_count = 2
            self.toggle_animation(False)
            imgui.open_popup("Export USD##export-usd-popup")

        imgui.set_next_window_size(570, 0)
        if imgui.begin_popup_modal(
            "Export USD##export-usd-popup",
            flags=imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_MOVE,
        )[0]:
            if self._export_usd_popup_open:
                # Export selection tree.
                def tree(nodes):
                    # Nodes GUI
                    for r in nodes:
                        # Skip nodes that shouldn't appear in the hierarchy.
                        if not r.show_in_hierarchy:
                            continue

                        # Visibility
                        curr_enabled = r.export_usd_enabled
                        if not curr_enabled:
                            imgui.push_style_color(imgui.COLOR_TEXT, 1.0, 1.0, 1.0, 0.4)

                        # Title
                        imgui.push_font(self.custom_font)
                        imgui.push_style_var(imgui.STYLE_FRAME_PADDING, (0, 2))

                        flags = imgui.TREE_NODE_OPEN_ON_ARROW | imgui.TREE_NODE_FRAME_PADDING
                        if r.export_usd_expanded:
                            flags |= imgui.TREE_NODE_DEFAULT_OPEN
                        if not any(c.show_in_hierarchy for c in r.nodes):
                            flags |= imgui.TREE_NODE_LEAF
                        r.export_usd_expanded = imgui.tree_node(
                            "{} {}##tree_node_{}".format(r.icon, r.name, r.unique_name), flags
                        )

                        imgui.pop_style_var()
                        imgui.pop_font()

                        if r != self:
                            # Aligns checkbox to the right side of the window
                            # https://github.com/ocornut/imgui/issues/196
                            imgui.same_line(position=imgui.get_window_content_region_max().x - 25)
                            eu, enabled = imgui.checkbox("##enabled_r_{}".format(r.unique_name), r.export_usd_enabled)
                            if eu:
                                r.export_usd_enabled = enabled

                        if r.export_usd_expanded:
                            # Recursively render children nodes
                            tree(r.nodes)
                            imgui.tree_pop()

                        if not curr_enabled:
                            imgui.pop_style_color(1)

                imgui.begin_child(
                    "export_select", height=300, border=True, flags=imgui.WINDOW_ALWAYS_VERTICAL_SCROLLBAR
                )
                tree(self.scene.nodes)
                imgui.end_child()

                if self.export_usd_name is None:
                    self.export_usd_name = "frame_{:0>6}".format(self.scene.current_frame_id)

                # HACK: we need to set the focus twice when the modal is first opened for it to take effect.
                if self._modal_focus_count > 0:
                    self._modal_focus_count -= 1
                    imgui.set_keyboard_focus_here()
                _, self.export_usd_name = imgui.input_text(
                    "File name",
                    self.export_usd_name,
                    64,
                    imgui.INPUT_TEXT_AUTO_SELECT_ALL,
                )

                _, self.export_usd_directory = imgui.checkbox(
                    "Export as directory with textures", self.export_usd_directory
                )

                imgui.spacing()

                button_width = (imgui.get_content_region_available()[0] - imgui.get_style().item_spacing[0]) * 0.5

                # Style the cancel with a grey color
                imgui.push_style_color(imgui.COLOR_BUTTON, 0.5, 0.5, 0.5, 1.0)
                imgui.push_style_color(imgui.COLOR_BUTTON_ACTIVE, 0.6, 0.6, 0.6, 1.0)
                imgui.push_style_color(imgui.COLOR_BUTTON_HOVERED, 0.7, 0.7, 0.7, 1.0)

                if imgui.button("cancel", width=button_width):
                    imgui.close_current_popup()
                    self._export_usd_popup_open = False

                imgui.pop_style_color()
                imgui.pop_style_color()
                imgui.pop_style_color()

                imgui.same_line()
                if imgui.button("save", button_width):
                    if self.export_usd_name:
                        self.export_usd(
                            os.path.join(C.export_dir, "usd", f"{self.export_usd_name}"),
                            self.export_usd_directory,
                            True,
                        )
                    imgui.close_current_popup()
                    self._export_usd_popup_open = False

            else:
                imgui.close_current_popup()
            imgui.end_popup()

    def gui_exit(self):
        if self._exit_popup_open:
            imgui.open_popup("Exit##exit-popup")

        if imgui.begin_popup_modal("Exit##exit-popup", flags=imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_MOVE)[0]:
            if self._exit_popup_open:
                imgui.text("Are you sure you want to exit?")
                imgui.spacing()

                # Draw a cancel and exit button on the same line using the available space
                button_width = (imgui.get_content_region_available()[0] - imgui.get_style().item_spacing[0]) * 0.5

                # Style the cancel with a grey color
                imgui.push_style_color(imgui.COLOR_BUTTON, 0.5, 0.5, 0.5, 1.0)
                imgui.push_style_color(imgui.COLOR_BUTTON_ACTIVE, 0.6, 0.6, 0.6, 1.0)
                imgui.push_style_color(imgui.COLOR_BUTTON_HOVERED, 0.7, 0.7, 0.7, 1.0)

                if imgui.button("cancel", width=button_width):
                    imgui.close_current_popup()
                    self._exit_popup_open = False

                imgui.pop_style_color()
                imgui.pop_style_color()
                imgui.pop_style_color()

                imgui.same_line()
                if imgui.button("exit", button_width):
                    self.window.close()
            else:
                imgui.close_current_popup()
            imgui.end_popup()

    def _mouse_to_buffer(self, x: int, y: int):
        return int(x * self.wnd.pixel_ratio), int(self.wnd.buffer_height - (y * self.wnd.pixel_ratio))

    def _mouse_to_viewport(self, x: int, y: int, viewport: Viewport):
        x, y = self._mouse_to_buffer(x, y)
        x = x - viewport.extents[0]
        y = y - viewport.extents[1]
        return x, viewport.extents[3] - y

    def mesh_mouse_intersection(self, x: int, y: int):
        """Given an x/y screen coordinate, get the intersected object, triangle id, and xyz point in camera space"""
        x, y = self._mouse_to_buffer(x, y)
        point_world, obj_id, tri_id, instance_id = self.renderer.read_fragmap_at_pixel(x, y)

        if obj_id >= 0 and tri_id >= 0:
            node = self.scene.get_node_by_uid(obj_id)
            # Camera space to world space
            point_local = (np.linalg.inv(node.model_matrix) @ np.append(point_world, 1.0))[:-1]
            if isinstance(node, Meshes) or isinstance(node, Billboard) or isinstance(node, VariableTopologyMeshes):
                vert_id = node.closest_vertex_in_triangle(tri_id, point_local)
                bc_coords = node.get_bc_coords_from_points(tri_id, [point_local])
            elif isinstance(node, PointClouds):
                vert_id = tri_id
                bc_coords = np.array([1, 0, 0])
            else:
                vert_id = 0
                bc_coords = np.array([0, 0, 0])

            return MeshMouseIntersection(node, instance_id, tri_id, vert_id, point_world, point_local, bc_coords)

        return None

    def select_object(self, x: int, y: int):
        """Selects the object at pixel coordinates x, y, returns True if an object is selected"""
        mmi = self.mesh_mouse_intersection(x, y)
        if mmi is not None:
            node = mmi.node

            # Traverse all parents until one is found that is selectable
            while node.parent is not None:
                if node.is_selectable:
                    # If the selection is locked only allow selecting the locked object
                    if not self.lock_selection or node == self.scene.selected_object:
                        self.scene.select(node, mmi.node, mmi.instance_id, mmi.tri_id)
                        return True
                    else:
                        return False
                node = node.parent

        return False

    def center_view_on_selection(self):
        self.center_view_on_node(self.scene.selected_object, with_animation=True)

    def center_view_on_node(self, node, with_animation=False):
        if isinstance(node, Node):
            self.reset_camera()
            forward = self.scene.camera.forward
            bounds = node.current_bounds
            diag = np.linalg.norm(bounds[:, 0] - bounds[:, 1])
            dist = max(0.01, diag * 1.3)

            center = bounds.mean(-1)
            anim_time = 0.25 if with_animation else 0.0
            self.scene.camera.move_with_animation(center - forward * dist, center, anim_time)

    def resize(self, width: int, height: int):
        self.window_size = (width, height)
        self._resize_viewports()
        self.renderer.resize(width, height)
        self.imgui.resize(width, height)

    def files_dropped_event(self, x: int, y: int, paths):
        for path in paths:
            base, ext = os.path.splitext(path)
            if ext == ".obj" or ext == ".ply":
                import trimesh

                obj = trimesh.load(path)
                obj_mesh = Meshes(obj.vertices, obj.faces, name=os.path.basename(base))
                self.scene.add(obj_mesh)

    def key_event(self, key, action, modifiers):
        self.imgui.key_event(key, action, modifiers)

        # Handle keyboard shortcuts when the exit modal is open
        if action == self.wnd.keys.ACTION_PRESS and self._exit_popup_open:
            if key == self.wnd.keys.ENTER:
                self.window.close()
            elif key == self._exit_key:
                self._exit_popup_open = False
            return

        if action == self.wnd.keys.ACTION_PRESS and (self._screenshot_popup_open or self._export_usd_popup_open):
            if key == self.wnd.keys.ENTER:
                if self.screenshot_name:
                    self.take_screenshot(self.screenshot_name, self.screenshot_transparent)
                    self._screenshot_popup_open = False
                if self.export_usd_name:
                    self.export_usd(
                        os.path.join(C.export_dir, "usd", f"{self.export_usd_name}"), self.export_usd_directory, True
                    )
                    self._export_usd_popup_open = False
            elif key == self._exit_key:
                self._screenshot_popup_open = False
                self._export_usd_popup_open = False
            return

        if action == self.wnd.keys.ACTION_PRESS and self._go_to_frame_popup_open:
            if key == self._exit_key:
                self._go_to_frame_popup_open = False
            return

        if action == self.wnd.keys.ACTION_PRESS and not self.render_gui:
            if key == self._exit_key:
                self.render_gui = True
                return

        if self.imgui.io.want_capture_keyboard:
            return

        if action == self.wnd.keys.ACTION_PRESS:
            if key == self._go_to_frame_key:
                self._go_to_frame_popup_open = True

            if key == self._exit_key:
                self._exit_popup_open = True

            if key == self._pause_key:
                self.toggle_animation(not self.run_animations)

            elif key == self._next_frame_key:
                if not self.run_animations:
                    self.scene.next_frame()

            elif key == self._previous_frame_key:
                if not self.run_animations:
                    self.scene.previous_frame()

            elif key == self._shadow_key:
                self.shadows_enabled = not self.shadows_enabled

            elif key == self._show_camera_target_key:
                self.scene.camera_target.enabled = not self.scene.camera_target.enabled

            elif key == self._show_camera_trackball_key:
                self.scene.trackball.enabled = not self.scene.trackball.enabled

            elif key == self._orthographic_camera_key:
                self.reset_camera()
                self.scene.camera.is_ortho = not self.scene.camera.is_ortho

            elif key == self._center_view_on_selection_key:
                self.center_view_on_selection()

            elif key == self._mode_view_key:
                self.selected_mode = "view"

            elif key == self._mode_inspect_key:
                self.selected_mode = "inspect"

            elif key == self._dark_mode_key:
                self.scene.light_mode = "dark" if self.scene.light_mode != "dark" else "default"

            elif key == self._screenshot_key:
                self._screenshot_popup_just_opened = True

            elif key == self._save_cam_key:
                self.reset_camera()
                self.scene.camera.save_cam()
            elif key == self._load_cam_key:
                self.reset_camera()
                self.scene.camera.load_cam()

            elif key == self._visualize_key:
                self.visualize = not self.visualize

            elif key == self._lock_selection_key:
                self.lock_selection = not self.lock_selection

            # Pass onto selected object
            if isinstance(self.scene.gui_selected_object, Node):
                self.scene.gui_selected_object.key_event(key, self.wnd.keys)

        if action == self.wnd.keys.ACTION_RELEASE:
            pass

    def mouse_position_event(self, x, y, dx, dy):
        self._mouse_position = (x, y)
        self.imgui.mouse_position_event(x, y, dx, dy)

        if self.selected_mode == "inspect":
            self.mmi = self.mesh_mouse_intersection(x, y)

    def mouse_press_event(self, x: int, y: int, button: int):
        self.imgui.mouse_press_event(x, y, button)

        if not self.imgui_user_interacting:
            self._moving_camera_viewport = self.get_viewport_at_position(x, y)
            if self._moving_camera_viewport is None:
                return

            # Pan or rotate camera on middle click.
            if button == self._middle_mouse_button:
                self._pan_camera = self.wnd.modifiers.shift
                self._rotate_camera = not self.wnd.modifiers.shift

            # Rotate camera on left click.
            if button == self._left_mouse_button:
                self.reset_camera(self._moving_camera_viewport)
                self._rotate_camera = True
                self._pan_camera = False
                x, y = self._mouse_to_viewport(x, y, self._moving_camera_viewport)
                self._moving_camera_viewport.camera.rotate_start(x, y, *self._moving_camera_viewport.extents[2:])

            if button == self._right_mouse_button:
                self._pan_camera = True
                self._rotate_camera = False

            self._mouse_down_position = np.array([x, y])
            self._mouse_moved = False

    def mouse_release_event(self, x: int, y: int, button: int):
        self.imgui.mouse_release_event(x, y, button)

        self._pan_camera = False
        self._rotate_camera = False
        self._moving_camera_viewport = None

        if not self.imgui_user_interacting:
            # Select the mesh under the cursor on left click.
            if button == self._left_mouse_button and not self._mouse_moved:
                if not self.select_object(x, y):
                    # If selection is enabled and nothing was selected clear the previous selection
                    if not self.lock_selection:
                        self.scene.select(None)

    def mouse_drag_event(self, x: int, y: int, dx: int, dy: int):
        self.imgui.mouse_drag_event(x, y, dx, dy)

        if not self.imgui_user_interacting:
            if self._moving_camera_viewport is None:
                return

            if self._pan_camera:
                self.reset_camera(self._moving_camera_viewport)
                self._moving_camera_viewport.camera.pan(dx, dy)

            if self._rotate_camera:
                self.reset_camera(self._moving_camera_viewport)
                x, y = self._mouse_to_viewport(x, y, self._moving_camera_viewport)
                self._moving_camera_viewport.camera.rotate(x, y, dx, dy, *self._moving_camera_viewport.extents[2:])

            if (
                not self._mouse_moved
                and np.linalg.norm(np.array([x, y]) - self._mouse_down_position) > self._move_threshold
            ):
                self._mouse_moved = True

    def mouse_scroll_event(self, x_offset: float, y_offset: float):
        self.imgui.mouse_scroll_event(x_offset, y_offset)

        if not self.imgui_user_interacting:
            v = self.get_viewport_at_position(*self._mouse_position)
            if v:
                self.reset_camera(v)
                v.camera.dolly_zoom(np.sign(y_offset), self.wnd.modifiers.shift, self.wnd.modifiers.ctrl)

    def unicode_char_entered(self, char):
        self.imgui.unicode_char_entered(char)

    def save_current_frame_as_image(self, path, scale_factor=None, alpha=False):
        """Saves the current frame as an image to disk."""
        image = self.get_current_frame_as_image(alpha)
        if scale_factor is not None and scale_factor != 1.0:
            w = int(image.width * scale_factor)
            h = int(image.height * scale_factor)
            image = image.resize((w, h), Image.LANCZOS)
        image.save(path)

    def get_current_frame_as_image(self, alpha=False):
        return self.renderer.get_current_frame_as_image(alpha)

    def get_current_mask_image(self, color_map: Dict[int, Tuple[int, int, int]] = None, id_map: Dict[int, int] = None):
        return self.renderer.get_current_mask_image(color_map, id_map)

    def get_current_mask_ids(self, id_map: Dict[int, int] = None):
        return self.renderer.get_current_mask_ids(id_map)

    def get_current_depth_image(self):
        return self.renderer.get_current_depth_image(self.scene.camera)

    def on_close(self):
        """
        Clean up before destroying the window
        """
        # Shut down all streams.
        for s in self.scene.collect_nodes(obj_type=Streamable):
            s.stop()

        # Shut down server.
        if self.server is not None:
            self.server.close()

        # Clear the lru_cache on all shaders, we do this so that future instances of the viewer
        # have to recompile shaders with the current moderngl context.
        # See issue #12 https://github.com/eth-ait/aitviewer/issues/12
        clear_shader_cache()

    def take_screenshot(self, file_name=None, transparent_background=False):
        """Save the current frame to an image in the screenshots directory inside the export directory"""
        if file_name is None:
            file_name = "frame_{:0>6}.png".format(self.scene.current_frame_id)
        if not file_name.endswith(".png"):
            file_name += ".png"
        file_path = os.path.join(C.export_dir, "screenshots", file_name)
        self.export_frame(file_path, transparent_background=transparent_background)
        print(f"Screenshot saved to {file_path}")

    def export_frame(self, file_path, scale_factor: float = None, transparent_background=False):
        """Save the current frame to an image.
        :param file_path: the path where the image is saved.
        :param scale_factor: a scale factor used to scale the image. If None no scale factor is used and
          the image will have the same size as the viewer.
        """
        dir = os.path.dirname(file_path)
        if dir:
            os.makedirs(dir, exist_ok=True)

        # Store run_animation old value and set it to false.
        run_animations = self.run_animations
        self.run_animations = False

        # Render and save frame.
        self.render(0, 0, export=True, transparent_background=transparent_background)
        self.save_current_frame_as_image(file_path, scale_factor, transparent_background)

        # Restore run animation and update last frame rendered time.
        self.run_animations = run_animations
        self._last_frame_rendered_at = self.timer.time

    def export_usd(self, path: str, export_as_directory=False, verbose=False):
        from pxr import Usd, UsdGeom

        if export_as_directory:
            if path.endswith(".usd"):
                directory = path[:-4]
            else:
                directory = path
            os.makedirs(directory, exist_ok=True)
            path = os.path.join(directory, os.path.basename(directory) + ".usd")
        else:
            directory = None
            if not path.endswith(".usd"):
                path += ".usd"

        # Create a new file and setup scene parameters.
        stage = Usd.Stage.CreateNew(path)
        stage.SetStartTimeCode(1)
        stage.SetEndTimeCode(self.scene.n_frames)
        UsdGeom.SetStageUpAxis(stage, UsdGeom.Tokens.y)

        # Recursively export the scene.
        self.scene.export_usd(stage, "", directory, verbose)

        # Save file.
        stage.Save()

    def export_video(
        self,
        output_path,
        frame_dir=None,
        animation=True,
        animation_range=None,
        duration=10.0,
        frame=None,
        output_fps=60.0,
        rotate_camera=False,
        rotation_degrees=360.0,
        scale_factor=None,
        transparent=False,
        quality="medium",
        ensure_no_overwrite=True,
    ):
        # Load this module to reduce load time.
        import skvideo.io

        if rotate_camera and not isinstance(self.scene.camera, ViewerCamera):
            print("Cannot export a video with camera rotation while using a camera that is not a ViewerCamera")
            return

        if frame_dir is None and output_path is None:
            print("You must either specify a path where to render the images to or where to save the video to")
            return

        if frame_dir is not None:
            # We want to avoid overriding anything in an existing directory, so add suffixes.
            # We always do this, even if `ensure_no_overwrite` is False, because this might lead to unexpected videos
            # if data already exists in this directory.
            format_str = "{:0>4}"
            counter = 0
            candidate_dir = os.path.join(frame_dir, format_str.format(counter))
            while os.path.exists(candidate_dir):
                counter += 1
                candidate_dir = os.path.join(frame_dir, format_str.format(counter))
            frame_dir = os.path.abspath(candidate_dir)

            # The frame dir does not yet exist (we've made sure of it).
            os.makedirs(frame_dir)

        # Store the current camera and create a copy of it if required.
        saved_camera = self.scene.camera
        if rotate_camera:
            self.scene.camera = copy.deepcopy(self.scene.camera)

        # Remember viewer data.
        saved_curr_frame = self.scene.current_frame_id
        saved_run_animations = self.run_animations

        if animation:
            if animation_range is None:
                animation_range = [0, self.scene.n_frames - 1]

            if animation_range[1] < animation_range[0]:
                print("No frames rendered.")
                return

            # Compute duration of the animation at given playback speed
            animation_frames = (animation_range[1] - animation_range[0]) + 1
            duration = animation_frames / self.playback_fps

            # Setup viewer for rendering the animation
            self.run_animations = True
            self.scene.current_frame_id = animation_range[0]
            self._last_frame_rendered_at = 0
        else:
            self.run_animations = False
            self.scene.current_frame_id = frame

        # Compute exact number of frames if we have the playback fps is an exact multiple of the output fps
        if animation and np.fmod(self.playback_fps, output_fps) < 0.1:
            playback_count = int(np.round(self.playback_fps / output_fps))
            frames = (animation_range[1] - animation_range[0] + 1) // playback_count
            self.run_animations = False
            exact_playback = True
        else:
            frames = int(np.ceil(duration * output_fps))
            exact_playback = False

        dt = 1 / output_fps
        time = 0

        # Compute camera speed.
        az_delta = np.radians(rotation_degrees) / frames

        # Initialize video writer.
        if output_path is not None:
            path_video, path_gif, is_gif = get_video_paths(output_path, ensure_no_overwrite)
            pix_fmt = "yuva420p" if transparent else "yuv420p"
            outputdict = {
                "-pix_fmt": pix_fmt,
                "-vf": "pad=ceil(iw/2)*2:ceil(ih/2)*2",  # Avoid error when image res is not divisible by 2.
                "-r": str(output_fps),
            }

            if path_video.endswith("mp4"):
                quality_to_crf = {
                    "high": 23,
                    "medium": 28,
                    "low": 33,
                }
                # MP4 specific options
                outputdict.update(
                    {
                        "-c:v": "libx264",
                        "-preset": "slow",
                        "-profile:v": "high",
                        "-level:v": "4.0",
                        "-crf": str(quality_to_crf[quality]),
                    }
                )

            writer = skvideo.io.FFmpegWriter(
                path_video,
                inputdict={
                    "-framerate": str(output_fps),
                },
                outputdict=outputdict,
            )

        for i in tqdm(range(frames), desc="Rendering frames"):
            if rotate_camera:
                self.scene.camera.rotate_azimuth(az_delta)

            self.render(time, time + dt, export=True, transparent_background=transparent)
            img = self.get_current_frame_as_image(alpha=transparent)

            # Scale image by the scale factor.
            if scale_factor is not None and scale_factor != 1.0:
                w = int(img.width * scale_factor)
                h = int(img.height * scale_factor)
                img = img.resize((w, h), Image.LANCZOS)

            # Store the image to disk if a directory for frames was given.
            if frame_dir is not None:
                img_name = os.path.join(frame_dir, "frame_{:0>6}.png".format(i))
                img.save(img_name)

            # Write the frame to the video writer.
            if output_path is not None:
                writer.writeFrame(np.array(img))

            if exact_playback:
                self.scene.current_frame_id = self.scene.current_frame_id + playback_count

            time += dt

        # Save the video.
        if output_path is not None:
            writer.close()
            if is_gif:
                # Convert to gif.
                video_to_gif(path_video, path_gif, remove=True)

                print(f"GIF saved to {os.path.abspath(path_gif)}")
            else:
                print(f"Video saved to {os.path.abspath(output_path)}")
        else:
            print(f"Frames saved to {os.path.abspath(frame_dir)}")

        # Reset viewer data.
        self.scene.camera = saved_camera
        self.scene.current_frame_id = saved_curr_frame
        self.run_animations = saved_run_animations
        self._last_frame_rendered_at = self.timer.time



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import moderngl
import numpy as np
import trimesh
from moderngl_window.opengl.vao import VAO

from aitviewer.renderables.spheres import SpheresTrail
from aitviewer.scene.material import Material
from aitviewer.scene.node import Node
from aitviewer.shaders import (
    get_depth_only_program,
    get_fragmap_program,
    get_lines_instanced_program,
    get_outline_program,
    get_simple_unlit_program,
)
from aitviewer.utils import (
    compute_vertex_and_face_normals,
    set_lights_in_program,
    set_material_properties,
)
from aitviewer.utils.decorators import hooked
from aitviewer.utils.so3 import aa2rot_numpy as aa2rot

_CYLINDER_SECTORS = 8


def _create_disk(n_disks=1, radius=1.0, sectors=None, plane="xz"):
    """
    Create `n_disks` many disks centered at the origin with the given radius1.
    :param n_disks: How many disks to create.
    :param radius: Radius of the disks.
    :param sectors: How many lines to use to approximate the disk. Increasing this will lead to more vertex data.
    :param plane: In which plane to create the disk.
    :return: Vertices as a np array of shape (N, V, 3), and face data as a np array of shape (F, 3).
    """
    assert plane in ["xz", "xy", "yz"]
    sectors = sectors or _CYLINDER_SECTORS
    angle = 2 * np.pi / sectors

    c1 = "xyz".index(plane[0])
    c2 = "xyz".index(plane[1])
    c3 = "xyz".index("xyz".replace(plane[0], "").replace(plane[1], ""))

    # Vertex Data.
    vertices = np.zeros((n_disks, sectors + 1, 3))
    x = radius * np.cos(np.arange(sectors) * angle)
    y = radius * np.sin(np.arange(sectors) * angle)
    vertices[:, 1:, c1] = x
    vertices[:, 1:, c2] = y

    # Faces.
    faces = np.zeros((sectors, 3), dtype=np.int32)
    idxs = np.array(range(1, sectors + 1), dtype=np.int32)
    faces[:, 2] = idxs
    faces[:-1, 1] = idxs[1:]
    faces[-1, 1] = 1

    return {"vertices": vertices, "faces": faces}


def _create_cylinder_from_to(v1, v2, radius1=1.0, radius2=1.0, sectors=None):
    """
    Create cylinders from points v1 to v2.
    :param v1: A np array of shape (N, 3).
    :param v2: A np array of shape (N, 3).
    :param radius1: The radius at the bottom of the cylinder.
    :param radius2: The radius at the top of the cylinder.
    :param sectors: how many lines to use to approximate the bottom disk.
    :return: Vertices and normals as a np array of shape (N, V, 3) and face data in shape (F, 3), i.e. only one
      face array is created for all cylinders.
    """
    sectors = sectors or _CYLINDER_SECTORS
    n_cylinders = v1.shape[0]

    # Create bottom lid.
    bottom = _create_disk(n_disks=n_cylinders, radius=radius1, sectors=sectors)

    # We must also change the winding of the bottom triangles because we have backface culling enabled and
    # otherwise we wouldn't see the bottom lid even if the normals are correct.
    fs_bottom = bottom["faces"]
    fs_bottom[:, 1], fs_bottom[:, 2] = fs_bottom[:, 2], fs_bottom[:, 1].copy()

    # Create top lid.
    top = _create_disk(n_disks=n_cylinders, radius=radius2, sectors=sectors)
    p2 = np.zeros((n_cylinders, 3))
    p2[:, 1] = np.linalg.norm(v2 - v1, axis=-1)
    top["vertices"] = top["vertices"] + p2[:, np.newaxis]

    # Shift indices of top faces by how many vertices the bottom lid has.
    n_vertices = bottom["vertices"].shape[1]
    fs_top = top["faces"] + n_vertices

    # Create the faces that make up the coat between bottom and top lid.
    idxs_bot = np.array(range(1, sectors + 1), dtype=np.int32)
    idxs_top = idxs_bot + n_vertices
    fs_coat1 = np.zeros((sectors, 3), dtype=np.int32)
    fs_coat1[:, 0] = idxs_top
    fs_coat1[:-1, 1] = idxs_top[1:]
    fs_coat1[-1, 1] = idxs_top[0]
    fs_coat1[:, 2] = idxs_bot

    fs_coat2 = np.zeros((sectors, 3), dtype=np.int32)
    fs_coat2[:, 0] = fs_coat1[:, 1]
    fs_coat2[:-1, 1] = idxs_bot[1:]
    fs_coat2[-1, 1] = idxs_bot[0]
    fs_coat2[:, 2] = idxs_bot

    # Concatenate everything to create a single mesh.
    vs = np.concatenate([bottom["vertices"], top["vertices"]], axis=1)
    fs = np.concatenate([fs_bottom, fs_top, fs_coat1, fs_coat2], axis=0)

    # Compute smooth normals.
    vertex_faces = trimesh.Trimesh(vs[0], fs, process=False).vertex_faces
    ns, _ = compute_vertex_and_face_normals(vs[0:1], fs, vertex_faces, normalize=True)
    ns = np.repeat(ns, n_cylinders, axis=0)

    # Rotate cylinders to align the the given data.
    vs, ns = _rotate_cylinder_to(v2 - v1, vs, ns)

    # Translate cylinders to the given positions
    vs += v1[:, np.newaxis]

    return {"vertices": vs, "normals": ns, "faces": fs}


def _create_cone_from_to(v1, v2, radius=1.0, sectors=None):
    """
    Create a cone from points v1 to v2.
    :param v1: A np array of shape (N, 3).
    :param v2: A np array of shape (N, 3).
    :param radius: The radius for the disk of the cone.
    :param sectors: how many lines to use to approximate the bottom disk.
    :return: Vertices and normals as a np array of shape (N, V, 3) and face data in shape (F, 3), i.e. only one
      face array is created for all cones.
    """
    sectors = sectors or _CYLINDER_SECTORS
    n_cylinders = v1.shape[0]

    # Create bottom lid.
    bottom = _create_disk(n_cylinders, radius, sectors)

    # We must also change the winding of the bottom triangles because we have backface culling enabled and
    # otherwise we wouldn't see the bottom lid even if the normals are correct.
    fs_bottom = bottom["faces"]
    fs_bottom[:, 1], fs_bottom[:, 2] = fs_bottom[:, 2], fs_bottom[:, 1].copy()

    # Add the top position as a new vertex.
    p2 = np.zeros((n_cylinders, 3))
    p2[:, 1] = np.linalg.norm(v2 - v1, axis=-1)
    vs = np.concatenate([bottom["vertices"], p2[:, np.newaxis]], axis=1)
    n_vertices = vs.shape[1]
    idx_top = n_vertices - 1

    # Create the faces from the bottom lid to the top point.
    idxs_bot = np.array(range(1, sectors + 1), dtype=np.int32)
    fs_coat = np.zeros((sectors, 3), dtype=np.int32)
    fs_coat[:, 0] = idx_top
    fs_coat[:-1, 1] = idxs_bot[1:]
    fs_coat[-1, 1] = idxs_bot[0]
    fs_coat[:, 2] = idxs_bot

    # Concatenate everything to create a single mesh.
    fs = np.concatenate([fs_bottom, fs_coat], axis=0)

    # Compute smooth normals.
    vertex_faces = trimesh.Trimesh(vs[0], fs, process=False).vertex_faces
    ns, _ = compute_vertex_and_face_normals(vs[0:1], fs, vertex_faces, normalize=True)
    ns = np.repeat(ns, n_cylinders, axis=0)

    # Rotate cones to align the the given data.
    vs, ns = _rotate_cylinder_to(v2 - v1, vs, ns)

    # Translate cones to the given positions
    vs += v1[:, np.newaxis]

    return {"vertices": vs, "normals": ns, "faces": fs}


def _rotate_cylinder_to(target, vs, ns):
    """
    Rotate vertices and normals such that the main axis of the cylinder aligns with `target`.
    :param target: A np array of shape (N, 3) specifying the target direction for each cylinder given in `vs`.
    :param vs: A np array of shape (N, V, 3), i.e. the vertex data for each cylinder.
    :param ns: A np array of shape (N, V, 3), i.e. the normal data for each cylinder.
    :return: The rotated vertices and normals.
    """
    n_cylinders = vs.shape[0]
    s = np.array([[0.0, 1.0, 0.0]]).repeat(n_cylinders, axis=0)
    v = target / (np.linalg.norm(target, axis=1, keepdims=True) + 1e-16)
    a = np.cross(s, v, axis=1)
    dot = np.sum(s * v, axis=1)
    theta = np.arccos(dot)
    is_col = np.linalg.norm(a, axis=1) < 10e-6
    a[is_col] = np.array([1.0, 0.0, 0.0])
    theta[is_col] = 0.0
    theta[np.logical_and(dot < 0.0, is_col)] = np.pi
    a = a / np.linalg.norm(a, axis=1, keepdims=True)
    rot = aa2rot(a * theta[..., np.newaxis])
    vs = np.matmul(rot[:, np.newaxis], vs[..., np.newaxis]).squeeze(-1)
    ns = np.matmul(rot[:, np.newaxis], ns[..., np.newaxis]).squeeze(-1)
    return vs, ns


class Lines(Node):
    """Render lines as cylinders or cones. Can render approx. 600k lines at 40 fps."""

    def __init__(
        self,
        lines,
        r_base=0.01,
        r_tip=None,
        color=(0.0, 0.0, 1.0, 1.0),
        mode="line_strip",
        cast_shadow=True,
        **kwargs,
    ):
        """
        Initializer.
        :param lines: Set of 3D coordinates as a np array of shape (F, L, 3) or (L, 3).
        :param r_base: Thickness of the line.
        :param r_tip: If set, the thickness of the line will taper from r_base to r_tip. If set to 0.0 it will create
          a proper cone.
        :param color: Color of the line (4-tuple) or array of color (N_LINES, 4), one for each line.
        :param mode: 'lines' or 'line_strip'.
            'lines': a line is drawn from point 0 to 1, from 2 to 3, and so on, number of lines is L / 2.
            'line_strip': a line is drawn between all adjacent points, 0 to 1, 1 to 2 and so on, number of lines is L - 1.
        :param cast_shadow: If True the mesh casts a shadow on other objects.
        """
        if len(lines.shape) == 2:
            lines = lines[np.newaxis]
        assert len(lines.shape) == 3
        assert mode == "lines" or mode == "line_strip"
        if mode == "lines":
            assert lines.shape[1] % 2 == 0

        self._lines = lines
        self.mode = mode
        self.r_base = r_base
        self.r_tip = r_tip if r_tip is not None else r_base

        self.vertices, self.faces = self.get_mesh()
        self.n_lines = self.lines.shape[1] // 2 if mode == "lines" else self.lines.shape[1] - 1

        # Define a default material in case there is None.
        if isinstance(color, tuple) or len(color.shape) == 1:
            kwargs["material"] = kwargs.get("material", Material(color=color, ambient=0.2))
            self.line_colors = kwargs["material"].color
        else:
            assert (
                color.shape[1] == 4 and color.shape[0] == self.n_lines
            ), "Color must be a tuple of 4 values or a numpy array of shape (N_LINES, 4)"
            self.line_colors = color

        super(Lines, self).__init__(n_frames=self.lines.shape[0], **kwargs)

        self._need_upload = True
        self.draw_edges = False

        # Render passes.
        self.outline = True
        self.fragmap = True
        self.depth_prepass = True
        self.cast_shadow = cast_shadow

    @property
    def bounds(self):
        bounds = self.get_bounds(self.lines)
        r = max(self.r_base, self.r_tip)
        bounds[:, 0] -= r
        bounds[:, 1] += r
        return bounds

    @property
    def current_bounds(self):
        bounds = self.get_bounds(self.current_lines)
        r = max(self.r_base, self.r_tip)
        bounds[:, 0] -= r
        bounds[:, 1] += r
        return bounds

    @property
    def lines(self):
        return self._lines

    @lines.setter
    def lines(self, value):
        self._lines = value if len(value.shape) == 3 else value[np.newaxis]
        self.n_frames = self.lines.shape[0]
        self.redraw()

    @property
    def current_lines(self):
        idx = self.current_frame_id if self._lines.shape[0] > 1 else 0
        return self._lines[idx]

    @current_lines.setter
    def current_lines(self, lines):
        assert len(lines.shape) == 2
        idx = self.current_frame_id if self._lines.shape[0] > 1 else 0
        self._lines[idx] = lines
        self.redraw()

    @Node.color.setter
    def color(self, color):
        self.material.color = color
        self.line_colors = color
        self.redraw()

    @property
    def line_colors(self):
        if len(self._line_colors.shape) == 1:
            t = np.tile(np.array(self._line_colors), (self.n_lines, 1))
            return t
        else:
            return self._line_colors

    @line_colors.setter
    def line_colors(self, color):
        if isinstance(color, tuple):
            color = np.array(color)
        self._line_colors = color
        self.redraw()

    def on_frame_update(self):
        self.redraw()

    def redraw(self, **kwargs):
        self._need_upload = True

    @Node.once
    def make_renderable(self, ctx: moderngl.Context):
        self.prog = get_lines_instanced_program()

        vs_path = "lines_instanced_positions.vs.glsl"
        self.outline_program = get_outline_program(vs_path)
        self.depth_only_program = get_depth_only_program(vs_path)
        self.fragmap_program = get_fragmap_program(vs_path)

        self.vbo_vertices = ctx.buffer(self.vertices.astype("f4").tobytes())
        self.vbo_indices = ctx.buffer(self.faces.astype("i4").tobytes())
        self.vbo_instance_base = ctx.buffer(reserve=self.n_lines * 12)
        self.vbo_instance_tip = ctx.buffer(reserve=self.n_lines * 12)
        self.vbo_instance_color = ctx.buffer(reserve=self.n_lines * 16)

        self.vao = VAO()
        self.vao.buffer(self.vbo_vertices, "3f4", "in_position")
        self.vao.buffer(self.vbo_instance_base, "3f4/i", "instance_base")
        self.vao.buffer(self.vbo_instance_tip, "3f4/i", "instance_tip")
        self.vao.buffer(self.vbo_instance_color, "4f4/i", "instance_color")
        self.vao.index_buffer(self.vbo_indices)

    def _upload_buffers(self):
        if not self.is_renderable or not self._need_upload:
            return
        self._need_upload = False

        lines = self.current_lines
        if self.mode == "lines":
            v0s = lines[::2]
            v1s = lines[1::2]
        else:
            v0s = lines[:-1]
            v1s = lines[1:]

        self.vbo_instance_base.write(v0s.astype("f4").tobytes())
        self.vbo_instance_tip.write(v1s.astype("f4").tobytes())

        if len(self._line_colors.shape) > 1:
            self.vbo_instance_color.write(self._line_colors.astype("f4").tobytes())

    def render(self, camera, **kwargs):
        self._upload_buffers()

        prog = self.prog
        prog["r_base"] = self.r_base
        prog["r_tip"] = self.r_tip
        if len(self._line_colors.shape) == 1:
            prog["use_uniform_color"] = True
            prog["uniform_color"] = tuple(self.color)
        else:
            prog["use_uniform_color"] = False
        prog["draw_edges"].value = 1.0 if self.draw_edges else 0.0
        prog["win_size"].value = kwargs["window_size"]
        prog["clip_control"].value = (0, 0, 0)

        self.set_camera_matrices(prog, camera, **kwargs)
        set_lights_in_program(
            prog,
            kwargs["lights"],
            kwargs["shadows_enabled"],
            kwargs["ambient_strength"],
        )
        set_material_properties(prog, self.material)
        self.receive_shadow(prog, **kwargs)
        self.vao.render(prog, moderngl.TRIANGLES, instances=self.n_lines)

    def render_positions(self, prog):
        if self.is_renderable:
            self._upload_buffers()
            prog["r_base"] = self.r_base
            prog["r_tip"] = self.r_tip
            self.vao.render(prog, moderngl.TRIANGLES, instances=self.n_lines)

    def get_mesh(self):
        v0s = np.array([[0, 0, 0]], np.float32)
        v1s = np.array([[0, 0, 1]], np.float32)

        # If r_tip is below a certain threshold, we create a proper cone, i.e. with just a single vertex at the top.
        if self.r_tip < 1e-5:
            data = _create_cone_from_to(v0s, v1s, radius=1.0)
        else:
            data = _create_cylinder_from_to(v0s, v1s, radius1=1.0, radius2=1.0)

        return data["vertices"][0], data["faces"]

    @hooked
    def release(self):
        if self.is_renderable:
            self.vao.release()

    def update_frames(self, lines, frames):
        self.lines[frames] = lines
        self.redraw()

    def add_frames(self, lines):
        if len(lines.shape) == 2:
            lines = lines[np.newaxis]
        self.lines = np.append(self.lines, lines, axis=0)

    def remove_frames(self, frames):
        self.lines = np.delete(self.lines, frames, axis=0)
        self.redraw()


class Lines2D(Node):
    """Render 2D lines."""

    def __init__(
        self,
        lines,
        color=(0.0, 0.0, 1.0, 1.0),
        mode="line_strip",
        **kwargs,
    ):
        """
        Initializer.
        :param lines: Set of 3D coordinates as a np array of shape (F, L, 3) or (L, 3).
        :param color: Color of the line (4-tuple) or array of color (N_LINES, 4), one for each line.
        :param mode: 'lines' or 'line_strip'.
            'lines': a line is drawn from point 0 to 1, from 2 to 3, and so on, number of lines is L / 2.
            'line_strip': a line is drawn between all adjacent points, 0 to 1, 1 to 2 and so on, number of lines is L - 1.
        """
        if len(lines.shape) == 2:
            lines = lines[np.newaxis]
        assert len(lines.shape) == 3
        assert mode == "lines" or mode == "line_strip"
        if mode == "lines":
            assert lines.shape[1] % 2 == 0

        self._lines = lines
        self.mode = mode
        self.n_lines = self.lines.shape[1] // 2 if mode == "lines" else self.lines.shape[1] - 1
        self.is_renderable = False

        # Define a default material in case there is None.
        if isinstance(color, tuple) or len(color.shape) == 1:
            kwargs["material"] = kwargs.get("material", Material(color=color, ambient=0.2))
            self.line_colors = kwargs["material"].color
        else:
            assert (
                color.shape[1] == 4 and color.shape[0] == self.n_lines
            ), "Color must be a tuple of 4 values or a numpy array of shape (N_LINES, 4)"
            self.line_colors = color

        super(Lines2D, self).__init__(n_frames=self.lines.shape[0], **kwargs)

    @property
    def bounds(self):
        bounds = self.get_bounds(self.lines)
        return bounds

    @property
    def current_bounds(self):
        bounds = self.get_bounds(self.current_lines)
        return bounds

    @property
    def lines(self):
        return self._lines

    @lines.setter
    def lines(self, value):
        self._lines = value if len(value.shape) == 3 else value[np.newaxis]
        self.n_frames = self.lines.shape[0]
        self.redraw()

    @property
    def current_lines(self):
        idx = self.current_frame_id if self._lines.shape[0] > 1 else 0
        return self._lines[idx]

    @current_lines.setter
    def current_lines(self, lines):
        assert len(lines.shape) == 2
        idx = self.current_frame_id if self._lines.shape[0] > 1 else 0
        self._lines[idx] = lines
        self.redraw()

    @Node.color.setter
    def color(self, color):
        self.material.color = color
        self.line_colors = color
        self.redraw()

    @property
    def line_colors(self):
        if len(self._line_colors.shape) == 1:
            t = np.tile(np.array(self._line_colors), (self.n_lines, 1))
            return t
        else:
            return self._line_colors

    @line_colors.setter
    def line_colors(self, color):
        if isinstance(color, tuple):
            color = np.array(color)
        self._line_colors = color
        self.redraw()

    def on_frame_update(self):
        super().on_frame_update()
        self.redraw()

    def _get_vertices(self):
        vertices = self.current_lines
        if self.mode == "line_strip":
            expanded = np.zeros((self.n_lines * 2, 3))
            expanded[::2] = vertices[:-1]
            expanded[1::2] = vertices[1:]
            vertices = expanded
        return vertices

    def _get_colors(self):
        cols = self.line_colors
        doubled = np.zeros((self.n_lines * 2, 4))
        doubled[::2] = cols
        doubled[1::2] = cols
        return doubled

    def redraw(self, **kwargs):
        """Upload the current frame data to the GPU for rendering."""
        if not self.is_renderable:
            return

        self.vbo_vertices.write(self._get_vertices().astype("f4").tobytes())
        self.vbo_colors.write(self._get_colors().astype("f4").tobytes())

    @Node.once
    def make_renderable(self, ctx: moderngl.Context):
        self.prog = get_simple_unlit_program()

        self.vbo_vertices = ctx.buffer(self._get_vertices().astype("f4").tobytes(), dynamic=True)
        self.vbo_colors = ctx.buffer(self._get_colors().astype("f4").tobytes(), dynamic=True)
        self.vao = VAO("lines", mode=moderngl.LINES)
        self.vao.buffer(self.vbo_vertices, "3f", ["in_position"])
        self.vao.buffer(self.vbo_colors, "4f", ["in_color"])

    def render(self, camera, **kwargs):
        self.set_camera_matrices(self.prog, camera, **kwargs)
        self.vao.render(self.prog, vertices=self.n_lines * 2)

    @hooked
    def release(self):
        if self.is_renderable:
            self.vao.release()


class LinesTrail(Lines):
    """A sequence of lines that leave a trail, i.e. the past lines keep being rendered."""

    def __init__(self, lines, with_spheres=True, r_base=0.01, r_tip=None, color=(0.0, 0.0, 1.0, 1.0), **kwargs):
        if "mode" not in kwargs:
            kwargs["mode"] = "line_strip"
        super().__init__(lines, r_base, r_tip, color, **kwargs)
        self.n_frames = lines.shape[0]

        if with_spheres:
            spheres_trail = SpheresTrail(lines, radius=r_base * 4.0, color=color)
            self._add_node(spheres_trail, enabled=True, show_in_hierarchy=True)

    def on_frame_update(self):
        self.n_lines = self.current_frame_id
        super().on_frame_update()

    def make_renderable(self, ctx):
        super().make_renderable(ctx)
        self.on_frame_update()



from ..message import Message
from ..node import RemoteNode


class RemoteArrows(RemoteNode):
    MESSAGE_TYPE = Message.ARROWS

    def __init__(self, viewer, origins, tips, **kwargs):
        """
        This initializer takes a RemoteViewer object and all other arguments are forwarded
        to the Arrows constructor on the remote Viewer.
        See the Arrows class for more information about parameters.

        :param viewer: a RemoteViewer object that will be used to send this node.
        :param origins: Set of 3D coordinates of the base of the arrows as a np array of shape (N, B, 3) or (B, 3).
        :param tips: Set of 3D coordinates denoting the tip of the arrow (N, T, 3) or (T, 3).
        """
        super().__init__(
            viewer,
            origins=origins,
            tips=tips,
            **kwargs,
        )

    def add_frames(self, origins, tips):
        """
        Add frames to the remote Arrows node by adding new origins and tips.

        :param origins: Set of 3D coordinates of the base of the arrows as a np array of shape (N, B, 3) or (B, 3).
        :param tips: Set of 3D coordinates denoting the tip of the arrow (N, T, 3) or (T, 3).
        """
        return super().add_frames(origins=origins, tips=tips)

    def update_frames(self, origins, tips, frames):
        """
        Update frames of the remote Arrows node by updating the origins and tips.

        :param origins: Set of 3D coordinates of the base of the arrows as a np array of shape (N, B, 3) or (B, 3).
        :param tips: Set of 3D coordinates denoting the tip of the arrow (N, T, 3) or (T, 3).
        :param frames: a list of integer frame indices of size N or a single integer frame index.
        """
        return super().update_frames(origins=origins, tips=tips, frames=frames)



from ..message import Message
from ..node import RemoteNode


class RemoteSMPLSequence(RemoteNode):
    MESSAGE_TYPE = Message.SMPL

    def __init__(
        self,
        viewer,
        poses_body,
        **kwargs,
    ):
        """
        This initializer takes a RemoteViewer object and all other arguments are forwarded
        to the SMPLLayer or SMPLSequence classes.
        The keyword arguments "model_type", "gender" and "num_betas" are forwarded to SMPLLayer,
        while the remaining arguments are forwarded to SMPLSequence.

        :param viewer: a RemoteViewer object that will be used to send this node.
        :param poses_body: An np array of shape (N, N_JOINTS*3) containing the pose parameters of the
          body, i.e. without hands or face parameters.
        """
        super().__init__(
            viewer,
            poses_body=poses_body,
            **kwargs,
        )

    def add_frames(self, poses_body, poses_root=None, trans=None, betas=None):
        """
        Add frames to the remote SMPLSequence node by adding body poses.

        :param poses_body: An np array of shape (N, N_JOINTS*3) or (N_JOINTS) containing the
          pose parameters of the body, i.e. without hands or face parameters.
        :param poses_root: An optional np array of shape (N, 3) or (3) containing the global root orientation.
        :param betas: An optional np array of shape (N, N_BETAS) or (N_BETAS) containing the shape parameters.
        :param trans: An optional np array of shape (N, 3) or (3) containing a global translation that is
          applied to all joints and vertices.

        """
        return super().add_frames(poses_body=poses_body, poses_root=poses_root, trans=trans, betas=betas)

    def update_frames(self, poses_body, frames, poses_root=None, trans=None, betas=None):
        """
        Update frames of the remote SMPLSequence node by updating body poses.

        :param poses_body: An np array of shape (N, N_JOINTS*3) or (N_JOINTS) containing the
          pose parameters of the body, i.e. without hands or face parameters.
        :param poses_root: An optional np array of shape (N, 3) or (3) containing the global root orientation.
        :param betas: An optional np array of shape (N, N_BETAS) or (N_BETAS) containing the shape parameters.
        :param trans: An optional np array of shape (N, 3) or (3) containing a global translation that is
          applied to all joints and vertices.
        """

        return super().update_frames(
            poses_body=poses_body,
            frames=frames,
            poses_root=poses_root,
            trans=trans,
            betas=betas,
        )



from ..message import Message
from ..node import RemoteNode


class RemoteSpheres(RemoteNode):
    MESSAGE_TYPE = Message.SPHERES

    def __init__(self, viewer, positions, **kwargs):
        """
        This initializer takes a RemoteViewer object and all other arguments are forwarded
        to the Spheres constructor on the remote Viewer.
        See the Spheres class for more information about parameters.

        :param viewer: a RemoteViewer object that will be used to send this node.
        :param positions: A numpy array of shape (N, P, 3) or (P, 3) containing sphere positions.
        """
        super().__init__(
            viewer,
            positions=positions,
            **kwargs,
        )

    def add_frames(self, positions):
        """
        Add frames to the remote Spheres node by adding new sphere positions.

        :param positions: A numpy array of shape (N, P, 3) or (P, 3) containing sphere positions.
        """
        return super().add_frames(positions=positions)

    def update_frames(self, positions, frames):
        """
        Update frames of the remote Spheres node by updating the sphere positions.

        :param positions: A numpy array of shape (N, P, 3) or (P, 3) containing sphere positions.
        :param frames: a list of integer frame indices of size N or a single integer frame index.
        """
        return super().update_frames(positions=positions, frames=frames)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
from typing import Dict, Tuple

import numpy as np
from PIL.Image import Image

from aitviewer.viewer import Viewer


class HeadlessRenderer(Viewer):
    samples = 4
    window_type = "headless"

    def __init__(self, **kwargs):
        """
        Initializer.
        :param frame_dir: Where to save the frames to.
        :param kwargs: kwargs.
        """
        super().__init__(**kwargs)

    def run(self, frame_dir=None, video_dir=None, output_fps=60, **export_video_kwargs):
        """Same as self.save_video, kept for backward compatibility."""
        return self.save_video(frame_dir, video_dir, output_fps)

    def save_video(self, frame_dir=None, video_dir=None, output_fps=60, transparent=False, **export_video_kwargs):
        """
        Convenience method to run the headless rendering.
        :param frame_dir: Where to store the individual frames or None if you don't care.
        :param video_dir: If set will automatically generate a video from the images found in `frame_dir`. Must
          be specified if `frame_dir` is None.
        :param output_fps: Fps of the output video, if None uses 60fps as default
        :param transparent: Save video with a transparent background, this is only supported by ".webm" format
            and ignored otherwise.
        """
        self._init_scene()
        self.export_video(
            output_path=video_dir,
            frame_dir=frame_dir,
            animation=True,
            output_fps=output_fps,
            transparent=transparent,
            **export_video_kwargs,
        )

    def save_frame(self, file_path, scale_factor: float = None):
        """
        Run the headless viewer and render a single frame.
        :param file_path: the path where the image is saved.
        :param scale_factor: a scale factor used to scale the image. If None no scale factor is used and
          the image will have the same size as the viewer.
        """
        self._init_scene()
        self.export_frame(file_path, scale_factor)

    def save_depth(self, file_path):
        """
        Render and save the depth buffer, see 'get_depth()' for more information
        about the depth format.
        :param file_path: the path where the image is saved. The file is used by PIL to choose
        the file format, make sure that you use a format that supports 'F' mode PIL Images (e.g. tiff).
        """
        dir = os.path.dirname(file_path)
        if dir:
            os.makedirs(dir, exist_ok=True)
        self.get_depth().save(file_path)

    def save_mask(self, file_path, color_map: Dict[int, Tuple[int, int, int]] = None, id_map: Dict[int, int] = None):
        """
        Render and save a color mask as a 'RGB' PIL image.
        Each object in the mask has a uniform color computed from the Node UID (can be accessed from a node with 'node.uid').

        :param file_path: the path where the image is saved.
        :param color_map:
            if not None specifies the color to use for a given Node UID as a tuple (R, G, B) of integer values from 0 to 255.
            If None the color is computed as an hash of the Node UID instead.
        :param id_map:
            if not None the UIDs in the mask are mapped using this dictionary from Node UID to the specified ID.
            This mapping is applied before the color map (or before hashing if the color map is None).
        """
        dir = os.path.dirname(file_path)
        if dir:
            os.makedirs(dir, exist_ok=True)
        self.get_mask(color_map, id_map).save(file_path)

    def _render_frame(self):
        self._init_scene()

        # Store run_animation old value and set it to false.
        run_animations = self.run_animations
        self.run_animations = False

        # Render frame.
        self.render(0, 0, export=True)

        # Restore run animation and update last frame rendered time.
        self.run_animations = run_animations

    def get_frame(self) -> Image:
        """Render and return a single frame as a 'RGB' PIL image"""
        self._render_frame()
        return self.get_current_frame_as_image()

    def get_depth(self) -> Image:
        """
        Render and return the depth buffer as a 'F' PIL image.
        Depth is stored as the z coordinate in eye (view) space.
        Therefore values in the depth image represent the distance from the pixel to
        the plane passing through the camera and orthogonal to the view direction.
        Values are between the near and far plane distances of the camera used for rendering,
        everything outside this range is clipped by OpenGL.
        """
        self._render_frame()
        return self.get_current_depth_image()

    def get_mask_ids(self, id_map: Dict[int, int] = None) -> np.ndarray:
        """
        Return a mask as a numpy array of shape (height, width) and type np.uint32.
        Each element in the array is the UID of the node covering that pixel (can be accessed from a node with 'node.uid')
        or zero if not covered.

        :param id_map:
            if not None the UIDs in the mask are mapped using this dictionary to the specified ID.
            The final mask only contains the IDs specified in this mapping and zeros everywhere else.
        """
        self._render_frame()
        return self.get_current_mask_ids(id_map)

    def get_mask(self, color_map: Dict[int, Tuple[int, int, int]] = None, id_map: Dict[int, int] = None) -> Image:
        """
        Render and return a color mask as a 'RGB' PIL image.
        Each object in the mask has a uniform color computed from the Node UID (can be accessed from a node with 'node.uid').

        :param color_map:
            if not None specifies the color to use for a given Node UID as a tuple (R, G, B) of integer values from 0 to 255.
            If None the color is computed as an hash of the Node UID instead.
        :param id_map:
            if not None the UIDs in the mask are mapped using this dictionary from Node UID to the specified ID.
            This mapping is applied before the color map (or before hashing if the color map is None).
        """
        self._render_frame()
        return self.get_current_mask_image(color_map, id_map)



import moderngl
import numpy as np
import trimesh
from moderngl_window.opengl.vao import VAO

from aitviewer.renderables.meshes import Meshes
from aitviewer.scene.node import Node
from aitviewer.shaders import (
    get_marching_cubes_shader,
    get_smooth_lit_with_edges_program,
)
from aitviewer.utils.decorators import hooked
from aitviewer.utils.marching_cubes_table import TRIS_TABLE
from aitviewer.utils.utils import compute_vertex_and_face_normals_sparse


class Volume(Meshes):
    """
    A signed distance function volume that is meshed with marching cubes directly on the GPU.
    """

    # Constants for marching cubes shader. The product of BX, BY and BZ must be smaller than the maximum supported
    # group size in the opengl compute shader. This is usually at least 1024 on most implementations.
    BX = 8
    BY = 8
    BZ = 8
    COMPACT_GROUP_SIZE = 128

    def __init__(
        self, volume, size=(1, 1, 1), level=0.0, max_triangles=None, max_vertices=None, invert_normals=False, **kwargs
    ):
        """
        Initializer.

        volume:
        :param volume: np array of shape (Z, Y, X) of signed distance values.
        :param size: size of the volume in local units.
        :param level: the level set that is meshed using marching cubes.
        :param max_triangles: the maximum number of triangles allocated for the mesh. If none a default value of 15M is used.
        :param max_vertices: the maximum number of vertices allocated for the mesh. If none a default value of 15M is used.
        :param invert_normals: if true flips the normals of the output mesh.
        :param **kwargs: arguments forwarded to the Meshes constructor.
        """
        # Initialize empty mesh.
        super().__init__(np.zeros((1, 0, 3)), np.zeros((0, 3)), **kwargs)

        # Disable backface culling for this mesh.
        self.backface_culling = False

        # Compute constants for marching cubes.
        self.NZ = volume.shape[0]
        self.NY = volume.shape[1]
        self.NX = volume.shape[2]

        self.BLOCKS_X = (self.NX - 1 + (self.BX - 2)) // (self.BX - 1)
        self.BLOCKS_Y = (self.NY - 1 + (self.BY - 2)) // (self.BY - 1)
        self.BLOCKS_Z = (self.NZ - 1 + (self.BZ - 2)) // (self.BZ - 1)
        self.TOTAL_BLOCKS = self.BLOCKS_X * self.BLOCKS_Y * self.BLOCKS_Z

        if max_triangles is None:
            max_triangles = 1000 * 1000 * 15
        if max_vertices is None:
            max_vertices = 1000 * 1000 * 15
        self.MAX_VERTICES = max_vertices
        self.MAX_TRIANGLES = max_triangles

        # Store parameters.
        self._volume = volume
        self.level = level
        self.size = size
        self.spacing = size / (np.array(volume.shape[::-1]) - 1.0)
        self.invert_normals = invert_normals

        # Internal variables.
        self._need_update = True
        self._need_readback = True
        self._debug_gui = False

    @property
    def vertices(self):
        return self.current_vertices[np.newaxis]

    @vertices.setter
    def vertices(self, _):
        raise Exception(f"vertices cannot be set on {self.__class__.__name__} object")

    @property
    def current_vertices(self):
        self._readback_data()
        return self._vertices_cpu

    @current_vertices.setter
    def current_vertices(self, _):
        raise Exception(f"vertices cannot be set on {self.__class__.__name__} object")

    @property
    def vertex_normals(self):
        return self.current_vertex_normals[np.newaxis]

    @vertex_normals.setter
    def vertex_normals(self, _):
        raise Exception(f"vertex normals cannot be set on {self.__class__.__name__} object")

    @property
    def current_vertex_normals(self):
        self._readback_data()
        return self._normals_cpu

    @current_vertex_normals.setter
    def current_vertex_normals(self, _):
        raise Exception(f"vertex normals cannot be set on {self.__class__.__name__} object")

    @property
    def faces(self):
        self._readback_data()
        return self._faces_cpu

    @faces.setter
    def faces(self, _):
        raise Exception(f"faces cannot be set on {self.__class__.__name__} object")

    @property
    def face_normals(self):
        _vertex_faces_sparse = trimesh.geometry.index_sparse(self.vertices.shape[1], self.faces)
        return compute_vertex_and_face_normals_sparse(self.vertices, self.faces, _vertex_faces_sparse, normalize=True)[
            1
        ]

    @property
    def current_face_normals(self):
        _vertex_faces_sparse = trimesh.geometry.index_sparse(self.vertices.shape[1], self.faces)
        return compute_vertex_and_face_normals_sparse(
            self.current_vertices, self.faces, _vertex_faces_sparse, normalize=True
        )[1]

    @property
    def bounds(self):
        return self.current_bounds

    @property
    def current_bounds(self):
        min = np.array([0, 0, 0])
        max = self.size
        return self.get_bounds(np.vstack((min, max)))

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, volume):
        assert (
            self._volume.shape == volume.shape
        ), f"New volume shape {volume.shape} must match old volume shape {self._volume.shape}"
        self._volume = volume
        if self.is_renderable:
            self._texture3d.write(self._volume.astype(np.float32).tobytes())
        self.redraw()

    @hooked
    def redraw(self, **kwargs):
        self._need_update = True

    def _readback_data(self):
        if self._need_update:
            self._marching_cubes()

        if self._need_readback:
            num_verts = np.frombuffer(self._num_vertices.read(), dtype=np.uint32)[0]
            if num_verts > self.MAX_VERTICES:
                raise Exception(
                    f"Too many vertices (Got: {num_verts}, max is: {self.MAX_VERTICES}). Consider using a higher max_vertices parameter"
                )

            num_indices = np.frombuffer(self._draw_args.read(4), dtype=np.uint32)[0]
            if num_indices // 3 > self.MAX_TRIANGLES:
                raise Exception(
                    f"Too many triangles (Got: {num_indices // 3}, max is: {self.MAX_TRIANGLES}. Consider using a higher max_triangles parameter"
                )

            # This ensures that buffers used by the GPU are never mapped in host memory.
            # If we don't do this the buffers stay mapped on the host and GPU operations on
            # these buffer become significantly slower.
            def readback(buffer, size):
                readback_buffer = self.ctx.buffer(reserve=size)
                self.ctx.copy_buffer(readback_buffer, buffer, size)
                data = readback_buffer.read()
                readback_buffer.release()
                return data

            vert_bytes = readback(self._vertices, num_verts * 3 * 4)
            norm_bytes = readback(self._normals, num_verts * 3 * 4)
            tri_bytes = readback(self._triangles, num_indices * 4)

            self._vertices_cpu = np.frombuffer(vert_bytes, dtype=np.float32).reshape((num_verts, 3))
            self._normals_cpu = np.frombuffer(norm_bytes, dtype=np.float32).reshape((num_verts, 3))
            self._faces_cpu = np.frombuffer(tri_bytes, dtype=np.uint32).reshape((num_indices // 3, 3))

            self._need_readback = False

    def _marching_cubes(self):
        # Check surface blocks.
        with self.time_queries["check_surface"]:
            self._texture3d.bind_to_image(0, read=True, write=False)

            self._all_blocks.bind_to_storage_buffer(0)

            self.prog_check_surface["u_level"] = self.level
            self.prog_check_surface.run(group_x=self.BLOCKS_X, group_y=self.BLOCKS_Y, group_z=self.BLOCKS_Z)
            self.ctx.memory_barrier()

        # Compact surface blocks.
        with self.time_queries["compact"]:
            # Initialize number of groups in x, y, and z for dispatch.
            self._out_blocks.write(np.array([0, 1, 1], dtype=np.uint32))

            self._all_blocks.bind_to_storage_buffer(0)
            self._out_blocks.bind_to_storage_buffer(1)

            self.prog_compact["u_num_blocks"] = self.TOTAL_BLOCKS
            self.prog_compact.run(group_x=(self.TOTAL_BLOCKS + self.COMPACT_GROUP_SIZE - 1) // self.COMPACT_GROUP_SIZE)
            self.ctx.memory_barrier()

        # Compute marching cubes
        with self.time_queries["mc"]:
            # Initialize draw arguments: count, instance_count, first_index, base_vertex, base_instances.
            self._draw_args.write(np.array([0, 1, 0, 0, 0], dtype=np.uint32))
            self._num_vertices.write(b"\x00\x00\x00\x00")

            self._texture3d.bind_to_image(0, read=True, write=False)

            self._out_blocks.bind_to_storage_buffer(0)
            self._tris_table.bind_to_storage_buffer(1)
            self._vertices.bind_to_storage_buffer(2)
            self._normals.bind_to_storage_buffer(3)
            self._triangles.bind_to_storage_buffer(4)
            self._num_vertices.bind_to_storage_buffer(5)
            self._draw_args.bind_to_storage_buffer(6)

            self.prog_mc["u_level"] = self.level
            self.prog_mc["u_blocks"] = (self.BLOCKS_X, self.BLOCKS_Y)
            self.prog_mc["u_volume_spacing"] = tuple(self.spacing)
            self.prog_mc["u_max_indices"] = self.MAX_TRIANGLES * 3
            self.prog_mc["u_max_vertices"] = self.MAX_VERTICES
            self.prog_mc["u_invert_normals"] = self.invert_normals

            self.prog_mc.run_indirect(buffer=self._out_blocks)
            self.ctx.memory_barrier()

        self._need_update = False
        self._need_readback = True

    @Node.once
    def make_renderable(self, ctx: moderngl.Context):
        try:
            # Load shaders for drawing.
            vs = "lit_with_edges.glsl"
            positions_vs = "mesh_positions.vs.glsl"
            self._load_programs(vs, positions_vs)

            # Load compute shaders for marching cubes.
            BX = self.BX
            BY = self.BY
            BZ = self.BZ
            CS = self.COMPACT_GROUP_SIZE
            self.prog_check_surface = get_marching_cubes_shader("check_surface.cs.glsl", BX, BY, BZ, CS)
            self.prog_compact = get_marching_cubes_shader("compact.cs.glsl", BX, BY, BZ, CS)
            self.prog_mc = get_marching_cubes_shader("marching_cubes.cs.glsl", BX, BY, BZ, CS)

            # Upload lookup table.
            self._tris_table = ctx.buffer(TRIS_TABLE.tobytes())

            # Create compute shader buffers.
            self._texture3d = ctx.texture3d(
                (self.NX, self.NY, self.NZ), 1, self.volume.astype(np.float32).tobytes(), dtype="f4"
            )
            self._all_blocks = ctx.buffer(reserve=self.TOTAL_BLOCKS * 4, dynamic=True)
            self._out_blocks = ctx.buffer(reserve=self.TOTAL_BLOCKS * 4 + 12, dynamic=True)
            self._num_vertices = ctx.buffer(reserve=4, dynamic=True)
            self._draw_args = ctx.buffer(reserve=20, dynamic=True)

            self._triangles = ctx.buffer(reserve=self.MAX_TRIANGLES * 12, dynamic=True)
            self._vertices = ctx.buffer(reserve=self.MAX_VERTICES * 12, dynamic=True)
            self._normals = ctx.buffer(reserve=self.MAX_VERTICES * 12, dynamic=True)
            self._colors = ctx.buffer(reserve=16, dynamic=True)

            self.prog = get_smooth_lit_with_edges_program("lit_with_edges.glsl", 0)

            self.vao = VAO()
            self.vao.buffer(self._vertices, "3f4", "in_position")
            self.vao.buffer(self._normals, "3f4", "in_normal")
            self.vao.buffer(self._colors, "4f4", "in_color")
            self.vao.index_buffer(self._triangles)

            self.ctx = ctx

            self.time_queries = {
                "check_surface": ctx.query(time=True),
                "compact": ctx.query(time=True),
                "mc": ctx.query(time=True),
                "render": ctx.query(time=True),
            }
        except Exception as e:
            raise Exception(
                f"Failed to initialize Volume renderable. This renderable requires support for OpenGL 4.5 (not available on macOS and old GPUs).\nError:\n{e}"
            )

    def render(self, camera, **kwargs):
        if self._need_update:
            self._marching_cubes()
        prog = self._use_program(camera, **kwargs)
        with self.time_queries["render"]:
            self.vao.render_indirect(prog, self._draw_args, mode=moderngl.TRIANGLES, count=1)

    def render_positions(self, prog):
        if self._need_update:
            self._marching_cubes()
        self.vao.render_indirect(prog, self._draw_args, mode=moderngl.TRIANGLES, count=1)

    @hooked
    def release(self):
        # vao and vbos are released by Meshes release method.
        if self.is_renderable:
            self._tris_table.release()
            self._texture3d.release()
            self._all_blocks.release()
            self._out_blocks.release()
            self._num_vertices.release()
            self._draw_args.release()

    def gui(self, imgui):
        u, self.level = imgui.drag_float("Level", self.level, 2e-3, min_value=-1.0, max_value=1.0)
        if u:
            self.redraw()

        if self._debug_gui:
            # Draw debugging stats about marching cubes and rendering time.
            total = 0
            for k, v in self.time_queries.items():
                imgui.text(f"{k}: {v.elapsed * 1e-6:5.3f}ms")
                total += v.elapsed * 1e-6
            imgui.text(f"Total: {total: 5.3f}ms")

        super().gui(imgui)



import os
from tempfile import TemporaryDirectory

import numpy as np
import trimesh
from utils import (
    RESOURCE_DIR,
    noreference,
    reference,
    requires_ffmpeg,
    requires_smpl,
    viewer,
)

from aitviewer.configuration import CONFIG as C
from aitviewer.headless import HeadlessRenderer
from aitviewer.renderables.lines import Lines
from aitviewer.renderables.meshes import Meshes
from aitviewer.renderables.smpl import SMPLLayer, SMPLSequence
from aitviewer.renderables.spheres import Spheres
from aitviewer.scene.camera import OpenCVCamera, WeakPerspectiveCamera
from aitviewer.utils.so3 import aa2rot_numpy
from aitviewer.viewer import Viewer


@noreference
@requires_ffmpeg
def test_headless(viewer: HeadlessRenderer):
    sphere_positions = np.array(
        [
            [[0.0, 0, 0]],
            [[0.1, 0, 0]],
            [[0.2, 0, 0]],
            [[0.3, 0, 0]],
            [[0.4, 0, 0]],
        ],
        np.float32,
    )
    spheres = Spheres(sphere_positions, radius=0.05, color=(1.0, 0.0, 1.0, 1.0))

    viewer.playback_fps = 30.0
    viewer.scene.add(spheres)
    with TemporaryDirectory() as temp:
        mp4_path = os.path.join(temp, "video.mp4")
        gif_path = os.path.join(temp, "video.gif")
        frame_path = os.path.join(temp, "frames")

        viewer.save_video(frame_dir=frame_path, video_dir=mp4_path, output_fps=30)
        assert os.path.exists(os.path.join(temp, "video_0.mp4")), "MP4 file not found"
        frames = os.path.join(frame_path, "0000")
        assert os.path.exists(frames) and len(os.listdir(frames)) == sphere_positions.shape[0]

        viewer.save_video(frame_dir=frame_path, video_dir=gif_path, output_fps=30)
        assert os.path.exists(os.path.join(temp, "video_0.gif")), "GIF file not found"
        frames = os.path.join(frame_path, "0001")
        assert os.path.exists(frames) and len(os.listdir(frames)) == sphere_positions.shape[0]

        viewer.save_video(video_dir=mp4_path)
        assert os.path.exists(os.path.join(temp, "video_1.mp4")), "MP4 file not found"
        frames = os.path.join(frame_path, "0002")
        assert not os.path.exists(frames)


@reference()
@requires_smpl
def test_normals(viewer: Viewer):
    smpl_transparent = SMPLSequence.t_pose(
        SMPLLayer(model_type="smpl", gender="male", device=C.device),
        name="SMPL",
        position=np.array((-1, 0.0, 0.0)),
    )
    smpl_opaque = SMPLSequence.t_pose(
        SMPLLayer(model_type="smpl", gender="male", device=C.device),
        name="SMPL",
        position=np.array((1, 0.0, 0.0)),
    )

    smpl_transparent.mesh_seq.norm_coloring = True
    smpl_transparent.mesh_seq.color = smpl_transparent.mesh_seq.color[:3] + (0.5,)

    smpl_opaque.mesh_seq.norm_coloring = True
    viewer.scene.add(smpl_transparent, smpl_opaque)

    viewer.scene.camera.position = np.array([0.0, 0.5, 3.5])


@reference()
def test_empty(viewer: Viewer):
    viewer.scene.add(viewer.scene.floor)
    viewer.scene.add(viewer.scene.origin)
    viewer.scene.camera.position = np.array([0.2, 0.2, 0.2])
    viewer.scene.camera.target = np.array([0.0, 0.0, 0.0])


def add_cube(viewer: Viewer, pos):
    cube = trimesh.load(os.path.join(RESOURCE_DIR, "cube.obj"), process=False)
    cube_mesh = Meshes(cube.vertices, cube.faces, name="Cube", position=pos, flat_shading=True)
    viewer.scene.add(cube_mesh)


@reference()
def test_weak_perspective_camera(viewer: Viewer):
    add_cube(viewer, np.array((0, 0, -5)))
    camera = WeakPerspectiveCamera(
        np.array([0.36266993, 0.62252431]),
        np.array([0.0100468, 0.36682156]),
        1920,
        1080,
        viewer=viewer,
    )
    viewer.scene.origin.enabled = False
    viewer.set_temp_camera(camera)


@reference()
def test_opencv_camera(viewer: Viewer):
    add_cube(viewer, np.array((0, 0, -5)))
    extrinsics = np.array(
        [
            [1.0, 0.0, 0.0, -0.09347329],
            [0.0, -1.0, 0.0, 0.08958537],
            [0.0, 0.0, -1.0, 3.0267725],
        ],
        np.float32,
    )
    intrinsics = np.array(
        [
            [1.10851252e03, 0.00000000e00, 6.40000000e02],
            [0.00000000e00, 1.10851252e03, 3.60000000e02],
            [0.00000000e00, 0.00000000e00, 1.00000000e00],
        ],
        np.float32,
    )
    camera = OpenCVCamera(intrinsics, extrinsics, 1280, 720, viewer=viewer)
    viewer.scene.origin.enabled = False
    viewer.set_temp_camera(camera)


@reference()
def test_vertex_face_colors(viewer: Viewer):
    cube = trimesh.load(os.path.join(RESOURCE_DIR, "cube.obj"), process=False)

    face_colors = np.array(
        [
            [1, 0, 0, 1],
            [1, 1, 0, 1],
            [0, 1, 0, 1],
            [1, 0, 1, 1],
            [0, 0, 1, 1],
            [0, 1, 1, 1],
            [1, 0, 0, 1],
            [1, 1, 0, 1],
            [0, 1, 0, 1],
            [1, 0, 1, 1],
            [0, 0, 1, 1],
            [0, 1, 1, 1],
        ],
        dtype=np.float32,
    )

    per_face0 = Meshes(
        cube.vertices,
        cube.faces,
        name="Cube",
        face_colors=face_colors,
        position=[-3.0, 0.0, -3.0],
        flat_shading=True,
    )
    per_face1 = Meshes(
        cube.vertices,
        cube.faces,
        name="Cube",
        position=[-3.0, 0.0, 3.0],
        flat_shading=True,
    )
    per_face1.face_colors = face_colors
    per_face2 = Meshes(
        cube.vertices,
        cube.faces,
        name="Cube",
        face_colors=face_colors,
        position=[-3.0, 0.0, 0.0],
        flat_shading=True,
    )
    per_face2.face_colors = face_colors

    vertex_colors = np.array(
        [
            [1, 0, 0, 1],
            [1, 1, 0, 1],
            [0, 1, 0, 1],
            [1, 0, 1, 1],
            [0, 0, 1, 1],
            [0, 1, 1, 1],
            [0, 0, 0, 1],
            [1, 1, 1, 1],
        ],
        dtype=np.float32,
    )

    per_vertex0 = Meshes(
        cube.vertices,
        cube.faces,
        vertex_colors=vertex_colors,
        name="Cube",
        position=[3.0, 0.0, -3.0],
        flat_shading=True,
    )
    per_vertex1 = Meshes(
        cube.vertices,
        cube.faces,
        name="Cube",
        position=[3.0, 0.0, 0.0],
        flat_shading=True,
    )
    per_vertex1.vertex_colors = vertex_colors
    per_vertex2 = Meshes(
        cube.vertices,
        cube.faces,
        vertex_colors=vertex_colors,
        name="Cube",
        position=[3.0, 0.0, 3.0],
        flat_shading=True,
    )
    per_vertex2.vertex_colors = vertex_colors

    color = color = (0.8, 0.5, 0.1, 1)
    uniform1 = Meshes(
        cube.vertices,
        cube.faces,
        name="Cube",
        position=[0.0, 0.0, -3.0],
        color=color,
        flat_shading=True,
    )
    uniform2 = Meshes(
        cube.vertices,
        cube.faces,
        name="Cube",
        position=[0.0, 0.0, 0.0],
        color=color,
        flat_shading=True,
    )
    uniform2.color = color
    uniform3 = Meshes(
        cube.vertices,
        cube.faces,
        name="Cube",
        position=[0.0, 0.0, 3.0],
        flat_shading=True,
    )
    uniform3.color = color

    viewer.scene.camera.dolly_zoom(-100.0)
    viewer.scene.camera.rotate_azimuth_elevation(250, 250)
    viewer.scene.floor.enabled = False
    viewer.shadows_enabled = False
    viewer.scene.add(
        per_vertex0,
        per_vertex1,
        per_vertex2,
        per_face0,
        per_face1,
        per_face2,
        uniform1,
        uniform2,
        uniform3,
    )


@reference()
def test_instancing(viewer: Viewer):
    cube = trimesh.load(os.path.join(RESOURCE_DIR, "cube.obj"), process=False)
    p = np.linspace(np.array([-8, 0, 0]), np.array([8, 0, 0]), num=10)
    r = aa2rot_numpy(np.linspace([0, 0, 0], np.array([0, 2 * np.pi, 0]), num=10))
    s = np.linspace(0.4, 0.8, num=10)

    viewer.scene.camera.position = (10, 5, 7)
    viewer.scene.camera.target = (1.5, -1, -2)

    cube_mesh = Meshes.instanced(cube.vertices, cube.faces, positions=p, rotations=r, scales=s, flat_shading=True)
    viewer.scene.add(cube_mesh)


@reference()
def test_sphere_and_line_colors(viewer: Viewer):
    s_pos = np.linspace(np.array([-1, 0, 0]), np.array([1, 0, 0]), 10)
    ls_pos = np.linspace(np.array([-1, 0, 1]), np.array([1, 0, 1]), 10)
    ll_pos = np.linspace(np.array([-1, 0, -1]), np.array([1, 0, -1]), 10)

    s_cols = np.linspace(np.array([0, 0, 1, 1]), np.array([1, 0, 0, 1]), 10)
    ls_cols = np.linspace(np.array([0, 0, 1, 1]), np.array([1, 0, 0, 1]), 9)
    ll_cols = np.linspace(np.array([0, 0, 1, 1]), np.array([1, 0, 0, 1]), 5)

    viewer.scene.camera.position = (2, 2, 2)
    viewer.scene.camera.target = (0, 0, 0)

    viewer.scene.add(Spheres(s_pos, color=s_cols, radius=0.05))
    viewer.scene.add(Lines(ls_pos, color=ls_cols, r_base=0.05, mode="line_strip"))
    viewer.scene.add(Lines(ll_pos, color=ll_cols, r_base=0.05, mode="lines"))



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


def requires_ctx(func):
    def _decorator(self, *args, **kwargs):
        if self.ctx is None:
            raise ValueError(
                "This function needs access to an OpenGL context - did you forget to call make_renderable()?"
            )
        else:
            return func(self, *args, **kwargs)

    return _decorator


def default_to_current_frame(func):
    def _decorator(self, *args, **kwargs):
        if len(args) == 0:
            kwargs["frame_id"] = kwargs.get("frame_id", self._current_frame_id)
        return func(self, *args, **kwargs)

    return _decorator


class hooked:
    def __init__(self, fn):
        self.fn = fn

    def __set_name__(self, owner, name):
        func = self.fn

        def _decorator(self, *args, **kwargs):
            super_obj = super(owner, self)
            super_fn = getattr(super_obj, func.__name__)
            super_fn(*args, **kwargs)
            return func(self, *args, **kwargs)

        setattr(owner, name, _decorator)

    def __call__(self):
        assert (
            False
        ), "@hooked decorator object should never be called directly. This can happen if you apply this decorator to a function that is not a method."



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import moderngl
import numpy as np
from moderngl_window.opengl.vao import VAO

from aitviewer.scene.node import Node
from aitviewer.shaders import (
    get_fragmap_program,
    get_outline_program,
    get_simple_unlit_program,
)
from aitviewer.utils.decorators import hooked


class PointClouds(Node):
    """
    Draw a point clouds man!
    """

    def __init__(
        self,
        points,
        colors=None,
        point_size=5.0,
        color=(0.0, 0.0, 1.0, 1.0),
        z_up=False,
        icon="\u008c",
        pickable=True,
        **kwargs,
    ):
        """
        A sequence of point clouds. Each point cloud can have a varying number of points.
        :param points: Sequence of points (F, P, 3)
        :param colors: Sequence of Colors (F, C, 4) or None. If None, all points are colored according to `color`.
        :param point_size: Initial point size.
        :param color: Default color applied to all points of all frames if `colors` is not provided.
        :param z_up: If true the point cloud rotation matrix is initialized to convert from z-up to y-up data.
        """
        assert isinstance(points, list) or isinstance(points, np.ndarray)
        if colors is not None:
            assert len(colors) == len(points)

        self.points = points
        super(PointClouds, self).__init__(n_frames=len(self.points), color=color, icon=icon, **kwargs)

        self.fragmap = pickable

        # Render passes
        self.outline = True

        self.colors = colors
        self.point_size = point_size
        self.max_n_points = max([p.shape[0] for p in self.points])

        self.vao = VAO("points", mode=moderngl.POINTS)

        if z_up:
            self.rotation = np.matmul(np.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]]), self.rotation)

    @property
    def points(self):
        return self._points

    @points.setter
    def points(self, points):
        self._points = points
        self.n_frames = len(points)
        self.max_n_points = max([p.shape[0] for p in self.points])

    @property
    def colors(self):
        return self._colors

    @colors.setter
    def colors(self, colors):
        # Colors cannot be empty
        if colors is None:
            self._colors = [self.color]
        elif isinstance(colors, tuple) and len(colors) == 4:
            self._colors = [colors]
        elif isinstance(colors, list) or isinstance(colors, np.ndarray):
            assert len(colors) == self.n_frames
            assert colors[0].shape[-1] == 4
            self._colors = colors
        else:
            raise ValueError("Invalid colors: {}".format(colors))

    @Node.color.setter
    def color(self, color):
        """Update the color of the point cloud."""
        # This is a bit ill-defined because point clouds can have per-point colors, in which case we probably do
        # not want to override them with a single uniform color. We disallow this for now. The function is still useful
        # though to change the alpha, even if a point cloud has per-point colors.
        self.material.color = color
        if self.is_renderable:
            single_color = isinstance(self.colors[0], tuple) and len(self.colors[0]) == 4
            if single_color:
                self.colors = tuple(color)
            else:
                # Only update the colors if the alpha changed. Take any frame and any point to check if the alpha
                # changed because we always change every frame and every point.
                alpha_changed = abs(color[-1] - self.colors[0][0, -1]) > 0
                if alpha_changed:
                    for i in range(self.n_frames):
                        self.colors[i][..., -1] = color[-1]
        self.redraw()

    @property
    def current_points(self):
        idx = self.current_frame_id if len(self.points) > 1 else 0
        return self.points[idx]

    @property
    def current_colors(self):
        if len(self.colors) == 1:
            n_points = self.current_points.shape[0]
            return np.full((n_points, 4), self.colors[0])
        else:
            idx = self.current_frame_id if len(self.colors) > 1 else 0
            return self.colors[idx]

    @property
    def bounds(self):
        if len(self.points) == 0:
            return np.array([[0, 0], [0, 0], [0, 0]])

        bounds = np.array([[np.inf, np.NINF], [np.inf, np.NINF], [np.inf, np.NINF]])
        for i in range(len(self.points)):
            b = self.get_bounds(self.points[i])
            bounds[:, 0] = np.minimum(bounds[:, 0], b[:, 0])
            bounds[:, 1] = np.maximum(bounds[:, 1], b[:, 1])
        return bounds

    @property
    def current_bounds(self):
        return self.get_bounds(self.current_points)

    def on_frame_update(self):
        """Called whenever a new frame must be displayed."""
        super().on_frame_update()
        self.redraw()

    def redraw(self, **kwargs):
        """Upload the current frame data to the GPU for rendering."""
        if not self.is_renderable:
            return

        points = self.current_points.astype("f4").tobytes()
        colors = self.current_colors.astype("f4").tobytes()

        # Resize the VBOs if necessary. This can happen if new points are set after the `make_renderable` has been
        # called.
        if self.max_n_points * 3 * 4 > self.vbo_points.size:
            self.vbo_points.orphan(self.max_n_points * 3 * 4)
            self.vbo_colors.orphan(self.max_n_points * 4 * 4)

        self.vbo_points.write(points)
        self.vbo_colors.write(colors)

    def _clear_buffer(self):
        self.vbo_points.clear()
        self.vbo_colors.clear()

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx):
        ctx.point_size = self.point_size

        self.prog = get_simple_unlit_program()
        self.outline_program = get_outline_program("mesh_positions.vs.glsl")
        self.fragmap_program = get_fragmap_program("mesh_positions.vs.glsl")

        self.vbo_points = ctx.buffer(reserve=self.max_n_points * 3 * 4, dynamic=True)
        self.vbo_colors = ctx.buffer(reserve=self.max_n_points * 4 * 4, dynamic=True)
        self.vbo_points.write(self.current_points.astype("f4").tobytes())
        self.vbo_colors.write(self.current_colors.astype("f4").tobytes())
        self.vao.buffer(self.vbo_points, "3f", ["in_position"])
        self.vao.buffer(self.vbo_colors, "4f", ["in_color"])

        self.positions_vao = VAO("{}:positions".format(self.unique_name), mode=moderngl.POINTS)
        self.positions_vao.buffer(self.vbo_points, "3f", ["in_position"])

    def render(self, camera, **kwargs):
        self.set_camera_matrices(self.prog, camera, **kwargs)
        # Draw only as many points as we have set in the buffer.
        self.vao.render(self.prog, vertices=len(self.current_points))

    def render_positions(self, prog):
        if self.is_renderable:
            self.positions_vao.render(prog, vertices=len(self.current_points))

    @hooked
    def release(self):
        if self.is_renderable:
            self.vao.release()
            self.positions_vao.release(buffer=False)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from aitviewer.configuration import CONFIG as C
from aitviewer.models.supr import SUPRLayer
from aitviewer.renderables.smpl import SMPLSequence
from aitviewer.utils import to_numpy as c2c


class SUPRSequence(SMPLSequence):
    """
    Represents a temporal sequence of SMPL poses using the SUPR model.
    """

    def __init__(
        self,
        poses_body,
        smpl_layer,
        poses_root,
        betas=None,
        trans=None,
        device=None,
        include_root=True,
        normalize_root=False,
        is_rigged=True,
        show_joint_angles=False,
        z_up=False,
        post_fk_func=None,
        **kwargs,
    ):
        super(SUPRSequence, self).__init__(
            poses_body,
            smpl_layer,
            poses_root,
            betas,
            trans,
            device=device,
            include_root=include_root,
            normalize_root=normalize_root,
            is_rigged=is_rigged,
            show_joint_angles=show_joint_angles,
            z_up=z_up,
            post_fk_func=post_fk_func,
            **kwargs,
        )

    def fk(self, current_frame_only=False):
        """Get joints and/or vertices from the poses."""
        if current_frame_only:
            # Use current frame data.
            if self._edit_mode:
                poses_root = self._edit_pose[:3][None, :]
                poses_body = self._edit_pose[3:][None, :]
            else:
                poses_body = self.poses_body[self.current_frame_id][None, :]
                poses_root = self.poses_root[self.current_frame_id][None, :]

            trans = self.trans[self.current_frame_id][None, :]

            if self.betas.shape[0] == self.n_frames:
                betas = self.betas[self.current_frame_id][None, :]
            else:
                betas = self.betas
        else:
            # Use the whole sequence.
            if self._edit_mode:
                poses_root = self.poses_root.clone()
                poses_body = self.poses_body.clone()
                poses_root[self.current_frame_id] = self._edit_pose[:3]
                poses_body[self.current_frame_id] = self._edit_pose[3:]
            else:
                poses_body = self.poses_body
                poses_root = self.poses_root
            trans = self.trans
            betas = self.betas

        verts, joints = self.smpl_layer(
            poses_root=poses_root,
            poses_body=poses_body,
            betas=betas,
            trans=trans,
            normalize_root=self._normalize_root,
        )

        skeleton = self.smpl_layer.skeletons()["body"].T
        faces = self.smpl_layer.faces
        joints = joints[:, : skeleton.shape[0]]

        if current_frame_only:
            return c2c(verts)[0], c2c(joints)[0], c2c(faces), c2c(skeleton)
        else:
            return c2c(verts), c2c(joints), c2c(faces), c2c(skeleton)

    @classmethod
    def from_amass(
        cls,
        npz_data_path,
        start_frame=None,
        end_frame=None,
        sub_frames=None,
        log=True,
        fps_out=None,
        load_betas=False,
        z_up=True,
        **kwargs,
    ):
        raise ValueError("SUPR does not support loading from 3DPW.")

    @classmethod
    def from_3dpw(cls, **kwargs):
        raise ValueError("SUPR does not support loading from 3DPW.")

    @classmethod
    def t_pose(cls, model=None, betas=None, frames=1, **kwargs):
        """Creates a SMPL sequence whose single frame is a SMPL mesh in T-Pose."""

        if model is None:
            model = SUPRLayer(device=C.device)

        poses_body = np.zeros([frames, model.n_joints_body * 3])
        poses_root = np.zeros([frames, 3])
        return cls(
            poses_body=poses_body,
            smpl_layer=model,
            poses_root=poses_root,
            betas=betas,
            **kwargs,
        )
